from dataclasses import dataclass, field
from sssekai.__main__ import create_parser
from typing import Any, List
from argparse import ArgumentParser
from logging import basicConfig, getLogger
import sys

try:
    from GooeyEx import Gooey, GooeyParser
except ImportError as e:
    print("Please install sssekai[gui] to use the GUI")
    raise e


from sssekai.unity import sssekai_set_unity_version


@Gooey(
    show_preview_warning=False,
    program_name="sssekai",
    tabbed_groups=True,
    advanced=True,
    monospace_display=True,
    default_size=(800, 600),
    menu=[
        {
            "name": "Help",
            "items": [
                {
                    "type": "Link",
                    "menuTitle": "Project Wiki",
                    "url": "https://github.com/mos9527/sssekai/wiki",
                },
                {
                    "type": "Link",
                    "menuTitle": "GitHub",
                    "url": "https://github.com/mos9527/sssekai",
                },
            ],
        }
    ],
)
def __main__():
    from tqdm.std import tqdm as tqdm_c

    parser = create_parser(GooeyParser)
    args = parser.parse_args()

    class TqdmMutexStream:
        @staticmethod
        def write(__s):
            # Gooey[Ex] only reads output from stdout so we'd do that here.
            with tqdm_c.external_write_mode(file=sys.stdout, nolock=False):
                return sys.stdout.write(__s)

    basicConfig(
        level="DEBUG",
        format="%(asctime)s | %(levelname).1s | %(name)s %(message)s",
        datefmt="%H:%M:%S",
        stream=TqdmMutexStream,
    )
    # override unity version
    sssekai_set_unity_version(args.unity_version)
    if "func" in args:
        args.func(args)
    pass


if __name__ in {"__main__", "__gui__"}:
    __main__()



from io import BytesIO
from sssekai.crypto.AssetBundle import decrypt_iter

from . import UnityPy, sssekai_get_unity_version
from UnityPy import Environment


def load_assetbundle(file: BytesIO) -> Environment:
    UnityPy.config.FALLBACK_VERSION_WARNED = True
    UnityPy.config.FALLBACK_UNITY_VERSION = sssekai_get_unity_version()
    stream = BytesIO()
    for block in decrypt_iter(lambda nbytes: file.read(nbytes)):
        stream.write(block)
    return UnityPy.load(stream)



from sssekai.unity.AssetBundle import load_assetbundle
from UnityPy.enums import ClassIDType
from wannacri.usm import Usm
from os import path, remove, makedirs
from logging import getLogger

logger = getLogger(__name__)


def main_usmdemux(args):
    with open(args.infile, "rb") as f:
        env = load_assetbundle(f)
        datas = dict()
        for obj in env.objects:
            if obj.type in {ClassIDType.MonoBehaviour, ClassIDType.TextAsset}:
                data = obj.read()
                datas[data.m_Name] = data
        movieInfo = datas.get("MovieBundleBuildData", None)
        assert movieInfo, "Invalid AssetBundle. No MovieBundleBuildData found!"
        # movieInfo = movieInfo.read_typetree()
        usm_name = movieInfo["movieBundleDatas"][0]["usmFileName"][: -len(".bytes")]
        logger.info("USM: %s" % usm_name)
        usm_folder = path.join(args.outdir, usm_name)
        makedirs(usm_folder, exist_ok=True)
        usm_temp = path.join(usm_folder, usm_name + ".tmp")
        with open(usm_temp, "wb") as usmstream:
            for data in movieInfo["movieBundleDatas"]:
                usm = data["usmFileName"][: -len(".bytes")]
                usm = datas[usm]
                usmstream.write(usm.m_Script.encode("utf-8", "surrogateescape"))
        usm = Usm.open(usm_temp, encoding="shift-jis")
        usm.demux(path.join(args.outdir, usm_name), usm_name)
        remove(usm_temp)
        logger.info("Saved to %s/" % usm_folder)



from collections import defaultdict
from pickle import load, dump
from typing import BinaryIO, List, Mapping, Optional, Union, Tuple
from logging import getLogger
from dataclasses import dataclass, fields, is_dataclass, field
from functools import cached_property
from base64 import b64encode, b64decode
import json

logger = getLogger("sssekai.abcache")

REGION_JP_EN = {"jp", "en"}
REGION_ROW = {"tw", "kr", "cn"}
REGION_OPTIONS = REGION_JP_EN | REGION_ROW


def fromdict(klass: type, d: Union[Mapping, List], warn_missing_fields=True):
    """
    Recursively convert a dictionary to a dataclass instance, if applicable.
    https://stackoverflow.com/a/54769644
    https://gist.github.com/gatopeich/1efd3e1e4269e1e98fae9983bb914f22
    """

    while klass.__name__ == "Optional":
        klass = klass.__args__[0]  # Reduce Optional[T] -> T

    def ensure_iterable(d):
        if isinstance(d, Mapping):
            return d
        if hasattr(d, "__dict__"):
            return d.__dict__
        return dict()

    def check_field(key, fields):
        if not key in fields:
            if warn_missing_fields:
                logger.error(f"Field {key} of type {type(d[key])} not found in {klass}")
            return False
        return True

    if is_dataclass(klass):
        fieldtypes = {f.name: f.type for f in fields(klass)}
        d = ensure_iterable(d)
        return klass(
            **(
                {
                    f: fromdict(fieldtypes[f], d[f], warn_missing_fields)
                    for f in d
                    if check_field(f, fieldtypes)
                }
            )
        )
    if isinstance(d, list) and hasattr(klass, "__args__"):
        return [fromdict(klass.__args__[0], di, warn_missing_fields) for di in d]
    if isinstance(d, dict) and hasattr(klass, "__args__"):
        return {
            k: fromdict(klass.__args__[1], v, warn_missing_fields) for k, v in d.items()
        }
    return d


from requests import Session, Response, HTTPError
from msgpack import unpackb, packb

from sssekai import __version__, __version_tuple__
from sssekai.unity import sssekai_get_unity_version
from sssekai.crypto.APIManager import decrypt, encrypt, SEKAI_APIMANAGER_KEYSETS


@dataclass
class AbCacheConfig:
    app_region: str
    app_version: str
    app_platform: str
    app_hash: str

    ab_version: str = None  # Override AB version in url for ROW
    auth_credential: str = None  # JWT token for JP/EN, Base64 encoded JWT token for ROW

    version: Tuple[int, int, int] = (0, 0, 0)  # Internal versioning

    @property
    def auth_jwt(self) -> str:
        if self.app_region in REGION_JP_EN:
            return self.auth_credential
        else:
            return b64decode(self.auth_credential).decode()

    @property
    def auth_jwt_payload(self) -> dict:
        header, payload, signature = self.auth_jwt.split(".")
        return json.loads(b64decode(payload + "=="))

    @property
    def auth_userID(self) -> str:
        return self.auth_jwt_payload.get("userId", None)

    @property
    def auth_available(self):
        return self.auth_credential is not None

    @property
    def cache_version(self):
        return getattr(self, "version", (0, 0, 0))

    @property
    def cache_version_string(self):
        return "%s.%s.%s" % self.cache_version if self.cache_version else "pre-0.5.17"

    @property
    def is_up_to_date(self):
        return self.cache_version == __version_tuple__


@dataclass
class AbCacheEntry(dict):
    bundleName: str
    cacheFileName: str
    cacheDirectoryName: str
    hash: str
    category: str
    crc: int
    fileSize: int
    dependencies: List[str]
    paths: List[str]
    isBuiltin: bool
    # ROW only
    md5Hash: Optional[str] = None
    downloadPath: Optional[str] = None


@dataclass
class AbCacheIndex(dict):
    version: str
    os: Optional[str] = None  # Undefined in ROW
    bundles: Mapping[str, AbCacheEntry] = None


# DS for the game's APIs
# There're *plenty* differences across regions. The implementations here
#   - Contains all the fields, even if they're not used in all regions
#   - Only the fields that are common across regions are non Optional[T]
@dataclass
class SekaiAppVersion:
    systemProfile: str
    appVersion: str
    multiPlayVersion: str
    appVersionStatus: str
    assetVersion: str
    # These are sadly removed from 3.5.0 (JP)
    dataVersion: Optional[str] = None
    appHash: Optional[str] = None
    assetHash: Optional[str] = None


@dataclass
class SekaiSystemData:
    serverDate: int
    # These are all undefined in ROW
    timezone: Optional[str] = None
    profile: Optional[str] = None
    maintenanceStatus: Optional[str] = None
    appVersions: Optional[List[SekaiAppVersion]] = None

    @cached_property
    def app_version_dict(self):
        return {av.appVersion: av for av in self.appVersions or []}


@dataclass
class SekaiGameVersionData:
    profile: str
    assetbundleHostHash: str
    domain: str


@dataclass
class SekaiUserRegistrationData:
    userId: int
    # JP/EN only
    signature: Optional[str] = None
    platform: Optional[str] = None
    deviceModel: Optional[str] = None
    operatingSystem: Optional[str] = None
    registeredAt: Optional[int] = None


@dataclass
class SekaiUserData:
    userRegistration: SekaiUserRegistrationData
    # JP/EN Only
    credential: Optional[str] = None
    updatedResources: Optional[dict] = None
    # ROW
    sessionToken: Optional[str] = None

    @property
    def user_credentials(self):
        return self.credential or self.sessionToken


@dataclass
class SekaiUserAuthData:
    sessionToken: str
    appVersion: str
    multiPlayVersion: str
    dataVersion: str
    assetVersion: str

    # JP/EN only
    removeAssetVersion: Optional[str] = None
    assetHash: Optional[str] = None

    appVersionStatus: Optional[str] = None

    # JP/EN only
    isStreamingVirtualLiveForceOpenUser: Optional[bool] = None
    deviceId: Optional[str] = None
    updatedResources: Optional[dict] = None
    suiteMasterSplitPath: Optional[list] = None
    obtainedBondsRewardIds: Optional[list] = None  # JP 4.0+

    # ROW
    configs: Optional[List[dict]] = None


@dataclass
class SSSekaiDatabase:
    config: AbCacheConfig = None
    cached_headers: dict = field(default_factory=dict)

    sekai_user_data: SekaiUserData = None
    sekai_user_auth_data: SekaiUserAuthData = None
    sekai_abcache_index: AbCacheIndex = None
    sekai_system_data: SekaiSystemData = None
    sekai_gameversion_data: SekaiGameVersionData = None


class AbCacheBundleNotFoundError(Exception):
    bundleName: str

    def __init__(self, bundleName: str):
        self.bundleName = bundleName
        super().__init__("Bundle not found: %s" % bundleName)


class AbCacheUserNotAuthenticatedError(Exception):
    pass


class AbCache(Session):
    database: SSSekaiDatabase

    @property
    def SEKAI_API_ENDPOINT(self):
        match self.config.app_region:
            case "jp":
                return "https://production-game-api.sekai.colorfulpalette.org"
            case "en":
                return "https://n-production-game-api.sekai-en.com"
            case "tw":
                return "https://mk-zian-obt-cdn.bytedgame.com"
            case "kr":
                return "https://mkkorea-obt-prod01-cdn.bytedgame.com"
            case "cn":
                return "https://mkcn-prod-public-60001-1.dailygn.com"
            case _:
                raise NotImplementedError

    @property
    def SEKAI_API_GAMEVERSION_ENDPOINT(self):
        match self.config.app_region:
            case "jp":
                return "https://game-version.sekai.colorfulpalette.org"
            case "en":
                return "https://game-version.sekai-en.com"
            case _:
                raise NotImplementedError

    @property
    def SEKAI_ISSUE_ENDPOINT(self):
        match self.config.app_region:
            case "jp":
                return "https://issue.sekai.colorfulpalette.org"
            case _:
                raise NotImplementedError

    @property
    def SEKAI_AB_INFO_ENDPOINT(self):
        match self.config.app_region:
            case "jp":
                return f"https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle-info.sekai.colorfulpalette.org/api/version/{self.SEKAI_ASSET_VERSION}/os/{self.config.app_platform}"
            case "en":
                return f"https://assetbundle-info.sekai-en.com/api/version/{self.SEKAI_ASSET_VERSION}/os/{self.config.app_platform}"
            case "tw":  # NOTE: Android only
                return f"https://lf16-mkovscdn-sg.bytedgame.com/obj/sf-game-alisg/gdl_app_5245/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/online/android71/AssetBundleInfoNew.json"
            case "kr":  # NOTE: Android only
                return f"https://lf16-mkkr.bytedgame.com/obj/sf-game-alisg/gdl_app_292248/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/kr_online/android63/AssetBundleInfoNew.json"
            case "cn":  # NOTE: Android only
                # https://github.com/mos9527/sssekai/issues/28
                return f"https://lf3-mkcncdn-tos.dailygn.com/obj/sf-game-lf/gdl_app_5236/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/cn_online1/android91/AssetBundleInfoNew.json"
            case _:
                raise NotImplementedError

    @property
    def SEKAI_AB_ENDPOINT(self):
        match self.config.app_region:
            case "jp":
                return f"https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle.sekai.colorfulpalette.org/"
            case "en":
                return f"https://assetbundle.sekai-en.com/"
            case "tw":  # NOTE: Android only
                return f"https://lf16-mkovscdn-sg.bytedgame.com/obj/sf-game-alisg/gdl_app_5245/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/online/"
            case "kr":  # NOTE: Android only
                return f"https://lf16-mkkr.bytedgame.com/obj/sf-game-alisg/gdl_app_292248/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/kr_online/"
            case "cn":  # NOTE: Android only
                # https://github.com/mos9527/sssekai/issues/28
                return f"https://lf3-j1gamecdn-cn.dailygn.com/obj/sf-game-lf/gdl_app_5236/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/cn_online1/"
            case _:
                raise NotImplementedError

    def _update_request_headers(self):
        self.headers.update(
            {
                "Accept": "application/octet-stream",
                "Content-Type": "application/octet-stream",
                "Accept-Encoding": "deflate, gzip",
                "User-Agent": "UnityPlayer/%s" % sssekai_get_unity_version(),
                "X-Platform": self.config.app_platform.capitalize(),
                "X-DeviceModel": "sssekai/%s" % __version__,
                "X-OperatingSystem": self.config.app_platform.capitalize(),
                "X-Unity-Version": sssekai_get_unity_version(),
                "X-App-Version": self.SEKAI_APP_VERSION,
                "X-App-Hash": self.SEKAI_APP_HASH,
            }
        )

    @property
    def config(self):
        return self.database.config

    @config.setter
    def config(self, v):
        self.database.config = v

    @property
    def SEKAI_APP_VERSION(self):
        return self.config.app_version

    @property
    def SEKAI_APP_PLATFORM(self):
        return self.config.app_platform

    @property
    def SEKAI_APP_HASH(self):
        return self.config.app_hash

    @property
    def SEKAI_ASSET_VERSION(self):
        return self.database.sekai_system_data.app_version_dict[
            self.SEKAI_APP_VERSION
        ].assetVersion

    @property
    def SEKAI_AB_HASH(self):
        return self.database.sekai_user_auth_data.assetHash

    @property
    def SEKAI_AB_HOST_HASH(self):
        return self.database.sekai_gameversion_data.assetbundleHostHash

    @property
    def SEKAI_API_SYSTEM_DATA(self):
        return self.SEKAI_API_ENDPOINT + "/api/system"

    @property
    def SEKAI_API_USER(self):
        return self.SEKAI_API_ENDPOINT + "/api/user"

    @property
    def SEKAI_USERID(self):
        return self.config.auth_userID or (
            self.database.sekai_user_data.userRegistration.userId
            if self.database.sekai_user_data
            else None
        )

    @property
    def SEKAI_CREDENTIAL(self):
        """Keyed in SEKAI_API_USER responses. `accessToken` in ROW servers, `credential` in JP/EN."""
        return self.config.auth_credential or (
            self.database.sekai_user_data.credential
            if self.database.sekai_user_data
            else None
        )

    @property
    def is_authenticated(self):
        return self.database.sekai_user_auth_data is not None

    def raise_for_auth(self, *args):
        if not self.is_authenticated:
            raise AbCacheUserNotAuthenticatedError(*args)

    @property
    def SEKAI_API_USER_AUTH(self):
        if self.config.app_region in REGION_JP_EN:
            assert self.SEKAI_USERID, "User ID must be available"
            return f"{self.SEKAI_API_USER}/{self.SEKAI_USERID}/auth"
        else:
            # For ROW this is the endpoint that retrives SEKAI_API_USER level data as well
            return f"{self.SEKAI_API_USER}/auth"

    @property
    def SEKAI_API_USER_SUITE(self):
        self.raise_for_auth()
        return f"{self.SEKAI_API_ENDPOINT}/api/suite/user/{self.SEKAI_USERID}"

    @property
    def SEKAI_API_MASTER_SUITE_URLS(self):
        match self.config.app_region:
            case "jp":
                self.raise_for_auth()
                return [
                    f"{self.SEKAI_API_ENDPOINT}/api/{path}"
                    for path in self.database.sekai_user_auth_data.suiteMasterSplitPath
                ]
            case "en":
                self.raise_for_auth()
                return [
                    f"{self.SEKAI_API_ENDPOINT}/api/{path}"
                    for path in self.database.sekai_user_auth_data.suiteMasterSplitPath
                ]
            # NOTE: All ROW servers have hardcoded master data URLs in *one* file as of the time of writing.
            # NOTE: Some of the MessagePack schemas is also *omitted* in these endpoints. Good luck parsing them.
            case "tw":
                return [
                    "https://lf21-mkovscdn-sg.bytedgame.com/obj/sf-game-alisg/gdl_app_5245/MasterData/60001/master-data-138.info"
                ]
            case "kr":
                return [
                    "https://lf19-mkkr.bytedgame.com/obj/sf-game-alisg/gdl_app_292248/MasterData/60001/master-data-158.info"
                ]
            case "cn":
                # Verified by https://github.com/mos9527/sssekai/issues/24
                return [
                    "https://lf3-mkcncdn-tos.dailygn.com/obj/sf-game-lf/gdl_app_5236/MasterData/60001/master-data-13.info"
                ]

    @property
    def SEKAI_API_INFORMATION(self):
        return self.SEKAI_API_ENDPOINT + "/api/information"

    @property
    def SEKAI_AB_BASE_PATH(self):
        return f"{self.SEKAI_ASSET_VERSION}/{self.SEKAI_AB_HASH}/{self.config.app_platform}/"

    @property
    def SEKAI_ISSUE_SIGNATURE_ENDPOINT(self):
        return self.SEKAI_ISSUE_ENDPOINT + "/api/signature"

    @property
    def abcache_index(self):
        return self.database.sekai_abcache_index

    def request_packed(self, method: str, url: str, data: dict = None, **kwargs):
        """Send a request with packed data. Data will be packed and encrypted before sending.

        Args:
            method (str): HTTP method
            url (str): URL
            data (dict, optional): Payload data. Defaults to None.

        Returns:
            Response: Response object
        """
        self._update_request_headers()
        if data is not None:
            data = packb(data)
            data = encrypt(data, SEKAI_APIMANAGER_KEYSETS[self.config.app_region])
        resp = self.request(method=method, url=url, data=data, **kwargs)
        if 400 <= resp.status_code < 600:
            logger.error(f"{method} {url} {resp.status_code}")
            try:  # log the error message provided by the API.
                content = self.response_to_dict(resp)
                logger.error("response=%s" % content)
            except:
                logger.error("response=%s" % resp.content)
            logger.error(
                "Unexpected server-side error. Refer to https://github.com/mos9527/sssekai/wiki#debugging-abcache for more information."
            )
            resp.raise_for_status()
        return resp

    def response_to_dict(self, resp: Response, **kwargs):
        """Decrypt and unpack a response content to a dictionary.

        Args:
            resp (Response): Response object
            **kwargs: Additional arguments for MessagePack unpackb

        Returns:
            dict: Decrypted and unpacked data
        """
        data = decrypt(resp.content, SEKAI_APIMANAGER_KEYSETS[self.config.app_region])
        data = unpackb(data, **kwargs)
        return data

    def _update_user_auth_data(self):
        if self.config.auth_available:
            logger.debug("Updating user auth data")
            if self.config.app_region in REGION_JP_EN:
                resp = self.request_packed(
                    "PUT",
                    self.SEKAI_API_USER_AUTH,
                    data={"credential": self.SEKAI_CREDENTIAL, "deviceId": None},
                )
            else:
                resp = self.request_packed(
                    "POST",
                    self.SEKAI_API_USER_AUTH,
                    data={
                        "accessToken": self.SEKAI_CREDENTIAL,
                        "deviceId": None,
                        "userID": 0,
                    },
                )
            data = self.response_to_dict(resp)
            self.database.sekai_user_auth_data = fromdict(
                SekaiUserAuthData, data, False
            )
            if self.config.app_region in REGION_ROW:
                self.database.sekai_user_data = fromdict(SekaiUserData, data, False)
        else:
            if self.config.app_region in REGION_JP_EN:
                raise AbCacheUserNotAuthenticatedError(
                    "Auth config is required for accessing JP/EN game server content"
                )

    def _update_system_data(self):
        logger.debug("Updating system data")
        resp = self.request_packed("GET", self.SEKAI_API_SYSTEM_DATA)
        data = self.response_to_dict(resp)
        self.database.sekai_system_data = fromdict(SekaiSystemData, data)

    def _update_gameversion_data(self):
        if self.config.app_region in REGION_JP_EN:
            logger.debug("Updating game version data")
            resp = self.request_packed(
                "GET",
                self.SEKAI_API_GAMEVERSION_ENDPOINT
                + "/"
                + self.SEKAI_APP_VERSION
                + "/"
                + self.SEKAI_APP_HASH,
            )
            data = self.response_to_dict(resp)
            self.database.sekai_gameversion_data = fromdict(SekaiGameVersionData, data)

    def _update_abcache_index(self) -> AbCacheIndex:
        logger.debug("Updating Assetbundle index")
        resp = self.request_packed("GET", self.SEKAI_AB_INFO_ENDPOINT)
        data = self.response_to_dict(resp)
        self.database.sekai_abcache_index = fromdict(AbCacheIndex, data)
        return self.database.sekai_abcache_index

    def _update_signatures(self):
        """Only required for JP for the initial setup. After which the game would cache the signature/cookies."""
        if self.config.app_region in {"jp"}:
            logger.debug("Updating signatures")
            # XXX: Clear all cookies, it 403s otherwise. Does this trigger a rate limit/expiry check since our session would be tracked?
            self.headers["Cookie"] = ""
            self.cookies.clear()
            resp = self.request_packed("POST", self.SEKAI_ISSUE_SIGNATURE_ENDPOINT)
            self.headers["Cookie"] = resp.headers["Set-Cookie"]
            # HACK: Per RFC6265, Cookies should not be visible to subdomains since it's not set with Domain attribute (https://github.com/psf/requests/issues/2576)
            # But the other endpoints uses it nontheless. So we have to set it manually.

    def __init__(self, config: Optional[AbCacheConfig] = None):
        super().__init__()
        self.database = SSSekaiDatabase()
        self.config = config or AbCacheConfig(
            "unknown", "unknown", "unknown", "unknown"
        )
        self.config.version = __version_tuple__

    def update_client_headers(self):
        """Authenticate the user and update client headers WITHOUT updating the AssetBundle Index.

        NOTE:
            - For JP/EN servers, `auth_credential` NEED to be set or otherwise you won't be able to do anything.
            - For ROW servers, it's OK to leave them empty.
                - To access user-level data, however, you still NEED to be authenticated.
        """
        logger.debug("Updating metadata")
        logger.debug("Set config: %s" % self.config)
        try:
            self._update_signatures()
            self._update_system_data()
        except HTTPError as e:
            logger.warning("Attempting to update signatures: %s" % e)
            self._update_signatures()
            self._update_system_data()
        if self.config.app_region in REGION_JP_EN:
            version_newest = self.database.sekai_system_data.appVersions[-1]
            logger.debug("Newest App version: %s" % version_newest)
            if version_newest.appVersion != self.SEKAI_APP_VERSION:
                logger.warning("App version mismatch. This may cause issues.")
        self._update_gameversion_data()
        self._update_user_auth_data()
        if self.is_authenticated:
            self.headers.update(
                {
                    "X-Data-Version": self.database.sekai_user_auth_data.dataVersion,
                    "X-Asset-Version": self.database.sekai_user_auth_data.assetVersion,
                    "X-Session-Token": self.database.sekai_user_auth_data.sessionToken,
                }
            )
        return self.headers

    def update(self):
        """Update the cache with the latest AssetBundle Index data from the server.

        NOTE:
            AbCache should have been initialized with a valid config before calling this method.
            Please refer to `update_client_headers` for more information.
        """
        self.update_client_headers()
        self._update_abcache_index()
        if self.config.app_region in REGION_JP_EN:
            logger.info("Sekai AssetBundle version: %s" % self.SEKAI_ASSET_VERSION)
            logger.info("Sekai AssetBundle host hash: %s" % self.SEKAI_AB_HOST_HASH)

    def save(self, f: BinaryIO):
        self._update_request_headers()
        self.database.cached_headers = self.headers
        logger.info("Saving cache: %s" % self)
        dump(self.database, f)

    def load(self, f: BinaryIO):
        logger.info("Loading cache")
        self.database = load(f)
        if not self.database.config.is_up_to_date:
            logger.warning(
                "Cache is outdated (cache: %s, current: %s). It's HIGHLY recommended to update the cache to avoid issues."
                % (self.database.config.cache_version_string, __version__)
            )
        self.headers.update(self.database.cached_headers)
        logger.info("Cache loaded: %s" % self)

    @staticmethod
    def from_file(f: BinaryIO):
        cache = AbCache()
        cache.load(f)
        return cache

    def __repr__(self) -> str:
        return (
            f"AbCache(config={self.config} bundles={len(self.abcache_index.bundles)})"
        )

    def get_entry_by_bundle_name(self, bundleName: str) -> AbCacheEntry:
        return self.abcache_index.bundles.get(bundleName, None)

    def get_entry_download_url(self, entry: AbCacheEntry):
        if self.config.app_region in REGION_JP_EN:
            return self.SEKAI_AB_ENDPOINT + self.SEKAI_AB_BASE_PATH + entry.bundleName
        else:
            # https://github.com/mos9527/sssekai/issues/28
            return self.SEKAI_AB_ENDPOINT + entry.downloadPath + "/" + entry.bundleName

    def get_or_update_dependency_tree_flatten(self, bundleName: str, deps: set = None):
        """Get a flattened set of asset dependency bundle names (including itself) for a given entry.

        Returned order is topologically sorted.

        if 'deps' is provided, it will be used as the initial set of dependencies.
        """
        deps = deps or set()

        def dfs(bundleName):
            for dep in self.abcache_index.bundles[bundleName].dependencies:
                if dep not in deps:
                    dfs(dep)
                    deps.add(dep)

        dfs(bundleName)
        deps.add(bundleName)
        return deps



NAMES_CRC_TBL = {
    4021652092: "Parts/ArtMesh159_Skinning",
    1133065872: "Parts/ArtMesh159_Skinning2",
    1702931593: "Parts/ArtMesh161_Skinning",
    2396259621: "Parts/ArtMesh161_Skinning2",
    2394363786: "Parts/ArtMesh162_Skinning",
    389440292: "Parts/ArtMesh162_Skinning2",
    300985368: "Parts/ArtMesh163Skinning",
    1972486609: "Parts/ArtMesh164Skinning",
    3950300558: "Parts/ArtMesh164Skinning2",
    2208917453: "Parts/ArtMesh164_Skinning",
    1659557266: "Parts/ArtMesh165Skinning",
    938903439: "Parts/ArtMesh181_Skinning",
    1743115835: "Parts/ArtMesh181_Skinning2",
    3499729872: "Parts/ArtMesh21_Skinning",
    2631333242: "Parts/ArtMesh21_Skinning2",
    814342638: "Parts/ArtMesh325_Skinning",
    880554451: "Parts/ArtMesh327_Skinning",
    716505440: "Parts/ArtMesh329_Skinning",
    4176256596: "Parts/ArtMesh331_Skinning",
    2820643001: "Parts/PARTS",
    2831198909: "Parts/PARTS2",
    2251968337: "Parts/PARTS_10",
    523443947: "Parts/PARTS_13",
    2169998152: "Parts/PARTS_14",
    4132486110: "Parts/PARTS_15",
    408869618: "Parts/PARTS_17",
    4293326837: "Parts/PARTS_19",
    1125744062: "Parts/PARTS_22",
    874401064: "Parts/PARTS_23",
    2526871881: "Parts/PARTS_4",
    2022924389: "Parts/PARTS_6",
    1898224287: "Parts/PARTS_ARM_L",
    2334894076: "Parts/PARTS_ARM_R",
    2290070841: "Parts/PARTS_BALLOON",
    2761942943: "Parts/PARTS_BODY",
    3851622095: "Parts/PARTS_BODY_1",
    2090485621: "Parts/PARTS_BODY_2",
    347760994: "Parts/PARTS_CORE",
    3126818741: "Parts/PARTS_EAR_CAT",
    798560733: "Parts/PARTS_EYE",
    2173469883: "Parts/PARTS_EYE_EFFECT_L1",
    411424001: "Parts/PARTS_EYE_EFFECT_L2",
    1439546212: "Parts/PARTS_EYE_EFFECT_R1",
    3435457246: "Parts/PARTS_EYE_EFFECT_R2",
    1572268774: "Parts/PARTS_EYE_L",
    1407274922: "Parts/PARTS_EYE_L2",
    619085628: "Parts/PARTS_EYE_L3",
    2813974405: "Parts/PARTS_EYE_R",
    2275438709: "Parts/PARTS_EYE_R2",
    4037492963: "Parts/PARTS_EYE_R3",
    2049152842: "Parts/PARTS_FACE",
    2359702804: "Parts/PARTS_FACE2",
    4221633922: "Parts/PARTS_FACE3",
    2836404823: "Parts/PARTS_HAIR_BACK",
    4017318464: "Parts/PARTS_HAIR_BACK2",
    2557762262: "Parts/PARTS_HAIR_BACK3",
    2296643005: "Parts/PARTS_HAIR_FRONT",
    2945290945: "Parts/PARTS_HAIR_FRONT2",
    3632964183: "Parts/PARTS_HAIR_FRONT3",
    1689995882: "Parts/PARTS_HAIR_SIDE_L",
    2662627081: "Parts/PARTS_HAIR_SIDE_R",
    3636730545: "Parts/PARTS_HEAD",
    2792093965: "Parts/PARTS_HEAD2",
    1687227251: "Parts/PARTS_HEAD_EFFECT",
    3552382848: "Parts/PARTS_HEAD_EFFECT2",
    2763669270: "Parts/PARTS_HEAD_EFFECT3",
    3758823366: "Parts/PARTS_MOUTH",
    1748161183: "Parts/PARTS_MOUTH2",
    523624969: "Parts/PARTS_MOUTH3",
    4291526910: "Parts/PARTS_NOSE",
    784651957: "Parts/PARTS__C_O_R_E",
    3783974482: "Parts/Part",
    1220215918: "Parts/Part10",
    1069298936: "Parts/Part11",
    2796905794: "Parts/Part12",
    3518125524: "Parts/Part13",
    1339474039: "Parts/Part14",
    953282785: "Parts/Part15",
    2715328859: "Parts/Part16",
    3604976077: "Parts/Part17",
    1180732508: "Parts/Part18",
    828882122: "Parts/Part19",
    2672886795: "Parts/Part2",
    1670797229: "Parts/Part20",
    345073467: "Parts/Part21",
    2375562881: "Parts/Part22",
    4204725783: "Parts/Part23",
    1694208948: "Parts/Part24",
    2331374232: "Parts/Part26",
    4260545038: "Parts/Part27",
    1833820063: "Parts/Part28",
    441118473: "Parts/Part29",
    3897963677: "Parts/Part3",
    2056087276: "Parts/Part30",
    227169914: "Parts/Part31",
    2491615168: "Parts/Part32",
    3817092950: "Parts/Part33",
    2111873781: "Parts/Part34",
    182948451: "Parts/Part35",
    2481898457: "Parts/Part36",
    3840537423: "Parts/Part37",
    1951854302: "Parts/Part38",
    55696968: "Parts/Part39",
    1983029566: "Parts/Part4",
    902623275: "Parts/Part40",
    1120649405: "Parts/Part41",
    3686961415: "Parts/Part42",
    2898633105: "Parts/Part43",
    849425458: "Parts/Part44",
    1168508068: "Parts/Part45",
    3702478110: "Parts/Part46",
    2879939976: "Parts/Part47",
    991393817: "Parts/Part48",
    1276135567: "Parts/Part49",
    20287912: "Parts/Part5",
    752337258: "Parts/Part50",
    1540420092: "Parts/Part51",
    3269042246: "Parts/Part52",
    3051262160: "Parts/Part53",
    733611379: "Parts/Part54",
    2554118162: "Parts/Part6",
    4013682820: "Parts/Part7",
    2139417877: "Parts/Part8",
    142859651: "Parts/Part9",
    2634930170: "Parts/PartAccessory",
    3724361956: "Parts/PartArm",
    181549335: "Parts/PartArmL",
    2567760371: "Parts/PartArmL2",
    4041016436: "Parts/PartArmR",
    1296946732: "Parts/PartArmR2",
    977987258: "Parts/PartArmR3",
    3536744278: "Parts/PartArmRibbon",
    2725054883: "Parts/PartBalloon",
    2393564822: "Parts/PartBlazer",
    2347293104: "Parts/PartBody",
    1211544802: "Parts/PartBody1",
    2371943827: "Parts/PartBody1_1",
    342469673: "Parts/PartBody1_2",
    1668193471: "Parts/PartBody1_3",
    3510625624: "Parts/PartBody2",
    2788742606: "Parts/PartBody3",
    945569901: "Parts/PartBody4",
    2909410759: "Parts/PartCORE",
    527991563: "Parts/PartCORE2",
    1181069585: "Parts/PartCapeL",
    3161099378: "Parts/PartCapeR",
    1003312973: "Parts/PartCore",
    310595191: "Parts/PartCore2",
    3299033335: "Parts/PartCry",
    2390042313: "Parts/PartCryL",
    1954159530: "Parts/PartCryR",
    411115564: "Parts/PartDeco",
    2195240294: "Parts/PartDowncoat",
    2401653952: "Parts/PartEarCat",
    2439359934: "Parts/PartExpression",
    1899812842: "Parts/PartEyeAdd",
    2082874295: "Parts/PartEyeEffectL1",
    3845091853: "Parts/PartEyeEffectL2",
    2825329768: "Parts/PartEyeEffectR1",
    829312466: "Parts/PartEyeEffectR2",
    1094503337: "Parts/PartEyeL",
    3047070090: "Parts/PartEyeL2",
    3140744906: "Parts/PartEyeR",
    1642047061: "Parts/PartEyeR2",
    383293123: "Parts/PartEyeR3",
    1431611749: "Parts/PartFace",
    660111219: "Parts/PartFace2",
    1348423653: "Parts/PartFace3",
    896714094: "Parts/PartFrill",
    3434591003: "Parts/PartHair",
    3031058490: "Parts/PartHairBack",
    3698183671: "Parts/PartHairBack2",
    1808830924: "Parts/PartHairBackL",
    2447355055: "Parts/PartHairBackR",
    1986500206: "Parts/PartHairBackSkinning1",
    4017014740: "Parts/PartHairBackSkinning2",
    2557081410: "Parts/PartHairBackSkinning3",
    101540577: "Parts/PartHairBackSkinning4",
    1896501879: "Parts/PartHairBackSkinning5",
    2443541283: "Parts/PartHairDeco",
    2157148866: "Parts/PartHairFront",
    1866436331: "Parts/PartHairFront2",
    4209349304: "Parts/PartHairSide",
    3751108641: "Parts/PartHairSide2",
    1747467290: "Parts/PartHairSideL",
    2452029817: "Parts/PartHairSideR",
    1524055430: "Parts/PartHairside",
    1041645371: "Parts/PartHandL",
    2883600662: "Parts/PartHandL2",
    3706155392: "Parts/PartHandL3",
    3289973336: "Parts/PartHandR",
    2141290185: "Parts/PartHandR2",
    145116767: "Parts/PartHandR3",
    2529352700: "Parts/PartHandR4",
    2108841151: "Parts/PartHand_L",
    1098293836: "Parts/PartHand_L2",
    2277335516: "Parts/PartHand_R",
    259781104: "Parts/PartHareBack",
    3618148921: "Parts/PartHat",
    4155712670: "Parts/PartHead",
    227914602: "Parts/PartHead2",
    1159802403: "Parts/PartHeadAccessory",
    1712963455: "Parts/PartHeadDeco",
    497937938: "Parts/PartHeadEffect",
    3916522155: "Parts/PartHeadEffect2",
    2400162126: "Parts/PartHeadEffect_03",
    1222988053: "Parts/PartHeadPhones",
    1523874884: "Parts/PartHeadRibbon",
    3309336036: "Parts/PartHighlightL",
    1062160519: "Parts/PartHighlightR",
    1688702150: "Parts/PartHood",
    2210588766: "Parts/PartKatyusha",
    3239288476: "Parts/PartLowerBody",
    3014432234: "Parts/PartMoon",
    1889238377: "Parts/PartMouth",
    785120396: "Parts/PartMouth2",
    3819404183: "Parts/PartNeckRibbon",
    3502030545: "Parts/PartNose",
    1735176321: "Parts/PartParker",
    3718539789: "Parts/PartPonL",
    665540462: "Parts/PartPonR",
    2884797086: "Parts/PartRibbon",
    2407603119: "Parts/PartRibon_L",
    1972250316: "Parts/PartRibon_R",
    1711001070: "Parts/PartShirt",
    3670282009: "Parts/PartSketch",
    430471656: "Parts/PartTear",
    60825605: "Parts/PartTearL",
    4189003110: "Parts/PartTearR",
    1527994312: "Parts/PartTemp",
    1627218647: "Parts/PartTie",
    2203288786: "Parts/PartTwintailL",
    2036119985: "Parts/PartTwintailR",
    2706143167: "Parts/PartUpperBody",
    51296898: "Parts/PartWork",
    799151642: "Parts/Part_CORE",
    1138891839: "Parts/PartlowerBody",
    309063465: "Parts/Partvehicle",
    1087288124: "Parameters/PARAMCloudExtendX",
    2063567466: "Parameters/PARAMRainbowExtend",
    2426468873: "Parameters/PARAM_ANGLE_X",
    3886533279: "Parameters/PARAM_ANGLE_Y",
    2125372197: "Parameters/PARAM_ANGLE_Z",
    1469304612: "Parameters/PARAM_ARM_L",
    3608623690: "Parameters/PARAM_ARM_L_1",
    1310616560: "Parameters/PARAM_ARM_L_2",
    957963110: "Parameters/PARAM_ARM_L_3",
    2810040005: "Parameters/PARAM_ARM_L_4",
    3497721427: "Parameters/PARAM_ARM_L_5",
    1232318441: "Parameters/PARAM_ARM_L_6",
    1047830399: "Parameters/PARAM_ARM_L_7",
    3816416182: "Parameters/PARAM_ARM_L_X",
    2912744007: "Parameters/PARAM_ARM_R",
    3249517616: "Parameters/PARAM_ARM_R_1",
    1487332746: "Parameters/PARAM_ARM_R_2",
    799135004: "Parameters/PARAM_ARM_R_3",
    2982494399: "Parameters/PARAM_ARM_R_4",
    3334631465: "Parameters/PARAM_ARM_R_5",
    1607147923: "Parameters/PARAM_ARM_R_6",
    684462341: "Parameters/PARAM_ARM_R_7",
    4123094476: "Parameters/PARAM_ARM_R_X",
    2814747804: "Parameters/PARAM_BALLOON_B_POS_X",
    3502412810: "Parameters/PARAM_BALLOON_B_POS_Y",
    910458965: "Parameters/PARAM_BALLOON_B_ROT_Z",
    2689469340: "Parameters/PARAM_BALLOON_B_X",
    168339544: "Parameters/PARAM_BALLOON_ROOT_Z",
    3223009362: "Parameters/PARAM_BALLOON_R_POS_X",
    3072067780: "Parameters/PARAM_BALLOON_R_POS_Y",
    1369054363: "Parameters/PARAM_BALLOON_R_ROT_Z",
    3160975596: "Parameters/PARAM_BALLOON_R_X",
    1967673382: "Parameters/PARAM_BALLOON_X",
    3256642728: "Parameters/PARAM_BALLOON_Y_POS_X",
    3038469182: "Parameters/PARAM_BALLOON_Y_POS_Y",
    1402816609: "Parameters/PARAM_BALLOON_Y_ROT_Z",
    1861975982: "Parameters/PARAM_BELT_SWING",
    728144516: "Parameters/PARAM_BODY_ANGLE_X",
    1549904402: "Parameters/PARAM_BODY_ANGLE_Y",
    3311991720: "Parameters/PARAM_BODY_ANGLE_Z",
    3387887330: "Parameters/PARAM_BODY_HANG_Z",
    2368228980: "Parameters/PARAM_BODY_SWING_A",
    337715150: "Parameters/PARAM_BODY_SWING_B",
    1663446872: "Parameters/PARAM_BODY_SWING_C",
    4249008891: "Parameters/PARAM_BODY_SWING_D",
    2319813229: "Parameters/PARAM_BODY_SWING_E",
    323804119: "Parameters/PARAM_BODY_SWING_F",
    1104289107: "Parameters/PARAM_BREATH",
    3371426049: "Parameters/PARAM_BREATH_BALLOON",
    4083107805: "Parameters/PARAM_BROW_L_ANGLE",
    905666941: "Parameters/PARAM_BROW_L_FORM",
    1978255271: "Parameters/PARAM_BROW_L_X",
    49215281: "Parameters/PARAM_BROW_L_Y",
    3300827214: "Parameters/PARAM_BROW_R_ANGLE",
    203883158: "Parameters/PARAM_BROW_R_FORM",
    1666271709: "Parameters/PARAM_BROW_R_X",
    341211467: "Parameters/PARAM_BROW_R_Y",
    2620500218: "Parameters/PARAM_BROW_WRINKLE_L",
    1715377561: "Parameters/PARAM_BROW_WRINKLE_R",
    850204031: "Parameters/PARAM_CHEEK",
    2000117621: "Parameters/PARAM_CHEEK_ANGRY",
    1225950358: "Parameters/PARAM_CLOTH_HUWA_A",
    2122070921: "Parameters/PARAM_EARSWING",
    1974843855: "Parameters/PARAM_EAR_L_FUWA",
    1282005540: "Parameters/PARAM_EAR_R_FUWA",
    3219453491: "Parameters/PARAM_EYEBALL_X",
    3370370725: "Parameters/PARAM_EYEBALL_Y",
    569471828: "Parameters/PARAM_EYE_BALL_X",
    1458979778: "Parameters/PARAM_EYE_BALL_Y",
    467979785: "Parameters/PARAM_EYE_FORM",
    1906573148: "Parameters/PARAM_EYE_L_FORM",
    2270921143: "Parameters/PARAM_EYE_L_OPEN",
    3178842395: "Parameters/PARAM_EYE_L_SMILE",
    1216322743: "Parameters/PARAM_EYE_R_FORM",
    3196576348: "Parameters/PARAM_EYE_R_OPEN",
    2325261960: "Parameters/PARAM_EYE_R_SMILE",
    3198949164: "Parameters/PARAM_EYE_SIZE",
    534080902: "Parameters/PARAM_EYE_TEARY",
    584930484: "Parameters/PARAM_FACE_SHADOW",
    201369778: "Parameters/PARAM_FINGER_L",
    4128216529: "Parameters/PARAM_FINGER_R",
    1889370645: "Parameters/PARAM_HAIR_BACK",
    2346709645: "Parameters/PARAM_HAIR_BACK_01",
    316056375: "Parameters/PARAM_HAIR_BACK_02",
    283660441: "Parameters/PARAM_HAIR_FRONT",
    3396622230: "Parameters/PARAM_HAIR_HUWA",
    1053972631: "Parameters/PARAM_HAIR_SIDE",
    979946251: "Parameters/PARAM_HAND_L_A",
    2741082801: "Parameters/PARAM_HAND_L_B",
    3563498023: "Parameters/PARAM_HAND_L_C",
    1241661316: "Parameters/PARAM_HAND_L_D",
    1023741714: "Parameters/PARAM_HAND_L_E",
    2752274088: "Parameters/PARAM_HAND_L_F",
    3540741694: "Parameters/PARAM_HAND_L_G",
    1135899567: "Parameters/PARAM_HAND_L_H",
    884163385: "Parameters/PARAM_HAND_L_I",
    2914652803: "Parameters/PARAM_HAND_L_J",
    3669828117: "Parameters/PARAM_HAND_L_K",
    751846769: "Parameters/PARAM_HAND_R_A",
    3050902731: "Parameters/PARAM_HAND_R_B",
    3269338205: "Parameters/PARAM_HAND_R_C",
    1555739134: "Parameters/PARAM_HAND_R_D",
    733839720: "Parameters/PARAM_HAND_R_E",
    2998194386: "Parameters/PARAM_HAND_R_F",
    3316899908: "Parameters/PARAM_HAND_R_G",
    1426914773: "Parameters/PARAM_HAND_R_H",
    571198787: "Parameters/PARAM_HAND_R_I",
    3137510649: "Parameters/PARAM_HAND_R_J",
    3422923887: "Parameters/PARAM_HAND_R_K",
    2004952014: "Parameters/PARAM_HEAD_ADJUST_L",
    2374901421: "Parameters/PARAM_HEAD_ADJUST_R",
    2437654298: "Parameters/PARAM_HEAD_HANG_Z",
    665245455: "Parameters/PARAM_HIGHLIGHT_VIEW",
    952355288: "Parameters/PARAM_LEG_L_Z",
    779831202: "Parameters/PARAM_LEG_R_Z",
    1485022100: "Parameters/PARAM_MOUTH_FORM",
    4007651115: "Parameters/PARAM_MOUTH_FORM_2",
    1049157407: "Parameters/PARAM_MOUTH_OPEN_Y",
    3686765564: "Parameters/PARAM_MOUTH_POSITION_Y",
    131783301: "Parameters/PARAM_MOUTH_SCALE_X",
    1893591571: "Parameters/PARAM_MOUTH_SCALE_Y",
    1975009948: "Parameters/PARAM_POSITION_X",
    46101002: "Parameters/PARAM_POSITION_Y",
    3459359932: "Parameters/PARAM_RIBBON_FUWA",
    1804253041: "Parameters/PARAM_RIBBON_SWING",
    2791945459: "Parameters/PARAM_ROTATION_Z",
    1545941616: "Parameters/PARAM_SHIRT_FUWA",
    4190776700: "Parameters/PARAM_SHIRT_SWING",
    113929410: "Parameters/PARAM_SHOULDER_L",
    4240789921: "Parameters/PARAM_SHOULDER_R",
    3868385143: "Parameters/PARAM_SIZE",
    1705512629: "Parameters/PARAM_SKIRT_FUWA",
    318609023: "Parameters/PARAM_SKIRT_SWING",
    1641055302: "Parameters/PARAM_SPECIAL_EYE",
    4127422515: "Parameters/PARAM_SWEAT_MOVE",
    2928453081: "Parameters/PARAM_SWEAT_ON",
    1543483275: "Parameters/PARAM_SWING_DRESS",
    4031979058: "Parameters/PARAM_SWING_DRESS_Y",
    2231506397: "Parameters/PARAM_SWING_RIBBON_1",
    470500455: "Parameters/PARAM_SWING_RIBBON_2",
    1795978481: "Parameters/PARAM_SWING_RIBBON_3",
    1488605431: "Parameters/PARAM_TEAR",
    1494750122: "Parameters/PARAM_TEAR_FROM",
    3638031514: "Parameters/PARAM_TEETH",
    2190374941: "Parameters/PARAM_TIE_SWING",
    3181879606: "Parameters/PARAM_UPPER_BODY",
    3750319138: "Parameters/PARAM_UPPER_BODY_Z",
    172220796: "Parameters/ParaBalloonFuwa",
    3410241356: "Parameters/ParaBalloonSwing",
    528322740: "Parameters/Param",
    3522937444: "Parameters/Param10",
    2801595122: "Parameters/Param11",
    1073062728: "Parameters/Param12",
    1223857118: "Parameters/Param13",
    3600156285: "Parameters/Param14",
    2710648555: "Parameters/Param15",
    949512017: "Parameters/Param16",
    1335842759: "Parameters/Param17",
    3743435350: "Parameters/Param18",
    2821159616: "Parameters/Param19",
    3603346440: "Parameters/Param2",
    4208365991: "Parameters/Param20",
    2379325745: "Parameters/Param21",
    349761675: "Parameters/Param22",
    1675608093: "Parameters/Param23",
    4256912830: "Parameters/Param24",
    2327602472: "Parameters/Param25",
    330642578: "Parameters/Param26",
    1689388036: "Parameters/Param27",
    2713814174: "Parameters/Param3",
    1161247105: "Parameters/Param3ParamArmL8",
    1067782461: "Parameters/Param4",
    1218585003: "Parameters/Param5",
    3517673489: "Parameters/Param6",
    2796306567: "Parameters/Param7",
    907238678: "Parameters/Param8",
    1091857792: "Parameters/Param9",
    2463981908: "Parameters/ParamAdjust",
    549326005: "Parameters/ParamAngleRotation24",
    1471757347: "Parameters/ParamAngleRotation25",
    3467668889: "Parameters/ParamAngleRotation26",
    3115801871: "Parameters/ParamAngleRotation27",
    688406686: "Parameters/ParamAngleRotation28",
    1578070024: "Parameters/ParamAngleRotation29",
    1053358573: "Parameters/ParamAngleRotation30",
    1238354299: "Parameters/ParamAngleRotation31",
    3502676161: "Parameters/ParamAngleRotation32",
    2814486615: "Parameters/ParamAngleRotation33",
    967130612: "Parameters/ParamAngleRotation34",
    1319243106: "Parameters/ParamAngleRotation35",
    3618331864: "Parameters/ParamAngleRotation36",
    2695654478: "Parameters/ParamAngleRotation37",
    1994697523: "Parameters/ParamAngleRotation44",
    31693733: "Parameters/ParamAngleRotation45",
    2565523999: "Parameters/ParamAngleRotation46",
    4025350793: "Parameters/ParamAngleRotation47",
    2136139544: "Parameters/ParamAngleRotation48",
    139843470: "Parameters/ParamAngleRotation49",
    1754422891: "Parameters/ParamAngleRotation50",
    397467875: "Parameters/ParamAngleX",
    1622659189: "Parameters/ParamAngleY",
    4190011855: "Parameters/ParamAngleZ",
    1035069756: "Parameters/ParamArm",
    3683193423: "Parameters/ParamArm0",
    1085069303: "Parameters/ParamArm01L1",
    3651503693: "Parameters/ParamArm01L2",
    2929899227: "Parameters/ParamArm01L3",
    2498619432: "Parameters/ParamArm01R1",
    233093522: "Parameters/ParamArm01R2",
    2061732100: "Parameters/ParamArm01R3",
    1122658734: "Parameters/ParamArm02L1",
    3689101332: "Parameters/ParamArm02L2",
    2900625538: "Parameters/ParamArm02L3",
    2527811185: "Parameters/ParamArm02R1",
    262277067: "Parameters/ParamArm02R2",
    2024093533: "Parameters/ParamArm02R3",
    1271936805: "Parameters/ParamArm0L",
    2984184390: "Parameters/ParamArm0R",
    2894996185: "Parameters/ParamArm1",
    1389037156: "Parameters/ParamArm1L",
    2831432455: "Parameters/ParamArm1R",
    898069347: "Parameters/ParamArm2",
    2045138343: "Parameters/ParamArm2L",
    2213112004: "Parameters/ParamArm2R",
    1115702261: "Parameters/ParamArm3",
    1627219174: "Parameters/ParamArm3L",
    564466456: "Parameters/ParamArmARSwitch",
    3951770818: "Parameters/ParamArmBMove",
    2938770683: "Parameters/ParamArmBRSwitch",
    1592314462: "Parameters/ParamArmFront",
    384977929: "Parameters/ParamArmL01",
    2415599027: "Parameters/ParamArmL02",
    3874360414: "Parameters/ParamArmL1",
    2028883422: "Parameters/ParamArmL10",
    2145869284: "Parameters/ParamArmL2",
    148925810: "Parameters/ParamArmL3",
    2525296849: "Parameters/ParamArmL4",
    3783510087: "Parameters/ParamArmL5",
    2022349309: "Parameters/ParamArmL6",
    260942187: "Parameters/ParamArmL7",
    2670900474: "Parameters/ParamArmL8",
    3895821420: "Parameters/ParamArmL9",
    1397086757: "Parameters/ParamArmLRibbonSwing1",
    3394014111: "Parameters/ParamArmLRibbonSwing2",
    3531658658: "Parameters/ParamArmLX",
    2777146676: "Parameters/ParamArmLY",
    4899443: "Parameters/ParamArmR01",
    2571342793: "Parameters/ParamArmR02",
    850342785: "Parameters/ParamArmR1",
    1851179940: "Parameters/ParamArmR10",
    2879808059: "Parameters/ParamArmR2",
    3701559981: "Parameters/ParamArmR3",
    1120257806: "Parameters/ParamArmR4",
    901969816: "Parameters/ParamArmR5",
    2899027490: "Parameters/ParamArmR6",
    3687618228: "Parameters/ParamArmR7",
    1265863461: "Parameters/ParamArmR8",
    1014283187: "Parameters/ParamArmR9",
    497315919: "Parameters/ParamArmRRibbonSwing1",
    2225938933: "Parameters/ParamArmRRibbonSwing2",
    113371773: "Parameters/ParamArmRX",
    1908857579: "Parameters/ParamArmRY",
    2280925350: "Parameters/ParamArmRise",
    775526053: "Parameters/ParamArmarillSwing",
    3158823164: "Parameters/ParamArmarillSwing2",
    4024589799: "Parameters/ParamBagSwingA",
    1995115613: "Parameters/ParamBagSwingB",
    32259275: "Parameters/ParamBagSwingC",
    763945921: "Parameters/ParamBalloonRibbonFuwa",
    2125669247: "Parameters/ParamBalloonRibbonSwing",
    3102943170: "Parameters/ParamBalloonSize",
    1862790838: "Parameters/ParamBalloonSize2",
    1456950351: "Parameters/ParamBalloonX",
    59120950: "Parameters/ParamBalloonYPosX",
    1063149258: "Parameters/ParamBeard",
    4203107101: "Parameters/ParamBeltFuwa",
    4043420464: "Parameters/ParamBeltSwing",
    1023250194: "Parameters/ParamBeltSwing2",
    2846399352: "Parameters/ParamBloodVessel",
    1257546180: "Parameters/ParamBoadUpDown",
    2257996805: "Parameters/ParamBoadUpDown2",
    369372051: "Parameters/ParamBoadUpDown2Body",
    3168167613: "Parameters/ParamBoadUpDownBody",
    2367297234: "Parameters/ParamBoadZ",
    1921325027: "Parameters/ParamBoadZ2",
    1590068145: "Parameters/ParamBoadZ2Body",
    3427274522: "Parameters/ParamBoadZBody",
    2933719285: "Parameters/ParamBody1",
    936661327: "Parameters/ParamBody2",
    2112179668: "Parameters/ParamBody2HangZ",
    1141202947: "Parameters/ParamBodyAngleX",
    855789717: "Parameters/ParamBodyAngleY",
    2852847919: "Parameters/ParamBodyAngleZ",
    2720197501: "Parameters/ParamBodyFrillHuwa",
    64735409: "Parameters/ParamBodyFrillSwing",
    1245006479: "Parameters/ParamBodyFuwa",
    2928948592: "Parameters/ParamBodyFuwaA",
    933061834: "Parameters/ParamBodyFuwaB",
    18333004: "Parameters/ParamBodyHangZ",
    2867477948: "Parameters/ParamBodyHuwa",
    295996177: "Parameters/ParamBodyHuwaA",
    2293095083: "Parameters/ParamBodyHuwaB",
    4289391165: "Parameters/ParamBodyHuwaC",
    1533977428: "Parameters/ParamBodyLegY",
    3261592302: "Parameters/ParamBodyLegZ",
    2899942910: "Parameters/ParamBodyLine",
    589872730: "Parameters/ParamBodyRibbonFuwa1",
    1994466024: "Parameters/ParamBodyRibbonSwing1",
    4024988498: "Parameters/ParamBodyRibbonSwing2",
    2565817284: "Parameters/ParamBodyRibbonSwing3",
    3538771839: "Parameters/ParamBodySwingA",
    1273237189: "Parameters/ParamBodySwingB",
    3099187545: "Parameters/ParamBodySwingB_01",
    565217507: "Parameters/ParamBodySwingB_02",
    1021525587: "Parameters/ParamBodySwingC",
    3901447549: "Parameters/ParamBodySwingC2",
    1597374790: "Parameters/ParamBodySwingCL",
    2771991589: "Parameters/ParamBodySwingCR",
    1399288064: "Parameters/ParamBodySwingC_L",
    2842184803: "Parameters/ParamBodySwingC_R",
    2726800368: "Parameters/ParamBodySwingD",
    276273025: "Parameters/ParamBodySwingDL",
    3933776610: "Parameters/ParamBodySwingDR",
    1445488517: "Parameters/ParamBodySwingD_L",
    2888260326: "Parameters/ParamBodySwingD_R",
    3581975398: "Parameters/ParamBodySwingE",
    1284098780: "Parameters/ParamBodySwingF",
    2516363576: "Parameters/ParamBodySwingF2",
    3808139694: "Parameters/ParamBodySwingF3",
    2090798093: "Parameters/ParamBodySwingF4",
    3595699470: "Parameters/ParamBodySwingFuwa",
    999225930: "Parameters/ParamBodySwingG",
    995804226: "Parameters/ParamBodySwingGL",
    3243637025: "Parameters/ParamBodySwingGR",
    2872174555: "Parameters/ParamBodySwingH",
    3694581581: "Parameters/ParamBodySwingI",
    1161791223: "Parameters/ParamBodySwingJ",
    842577505: "Parameters/ParamBodySwingK",
    2891719618: "Parameters/ParamBodySwingL",
    3680179028: "Parameters/ParamBodySwingM",
    1112687342: "Parameters/ParamBodySwingN",
    894792312: "Parameters/ParamBodySwingO",
    3093135245: "Parameters/ParamBodySwingP",
    3478802203: "Parameters/ParamBodySwingQ",
    1448287905: "Parameters/ParamBodySwingR",
    559164983: "Parameters/ParamBodySwingS",
    3207642004: "Parameters/ParamBodySwingT",
    3359083266: "Parameters/ParamBodySwingU",
    1363073720: "Parameters/ParamBodySwingV",
    2876944824: "Parameters/ParamBracelet1Fuwa",
    1475090613: "Parameters/ParamBracelet1Swing",
    3973761896: "Parameters/ParamBracelet2Fuwa",
    3514329627: "Parameters/ParamBracelet2Swing",
    1671053000: "Parameters/ParamBracelet4Fuwa",
    1732271208: "Parameters/ParamBraceletSwing",
    3823476906: "Parameters/ParamBreath",
    3120995542: "Parameters/ParamBroachSwing",
    321347094: "Parameters/ParamBrowLAngle",
    2550903895: "Parameters/ParamBrowLForm",
    1254050164: "Parameters/ParamBrowLWrinkle",
    2333223759: "Parameters/ParamBrowLWrinkle01",
    303790837: "Parameters/ParamBrowLWrinkle02",
    4199213146: "Parameters/ParamBrowLWrinkleL",
    2780107611: "Parameters/ParamBrowLX",
    3534889933: "Parameters/ParamBrowLY",
    721108477: "Parameters/ParamBrowRAngle",
    1205575092: "Parameters/ParamBrowRForm",
    596535917: "Parameters/ParamBrowRWrinkle",
    1911827588: "Parameters/ParamBrowRX",
    116595730: "Parameters/ParamBrowRY",
    1028878831: "Parameters/ParamBrowScale",
    2352716264: "Parameters/ParamBrowWrinkleL",
    1983161483: "Parameters/ParamBrowWrinkleR",
    1904893592: "Parameters/ParamBustFuriruFuwa",
    1817757842: "Parameters/ParamBustFuriruSwing",
    3288276072: "Parameters/ParamBustRibbonFuwa",
    164668746: "Parameters/ParamBustX",
    2128057820: "Parameters/ParamBustY",
    4080828015: "Parameters/ParamCaneSwitch",
    1390522448: "Parameters/ParamCardiganFuwa",
    4190431742: "Parameters/ParamCardiganSwing",
    730756834: "Parameters/ParamChainHuwaB",
    3712910823: "Parameters/ParamChainSwingA",
    610699091: "Parameters/ParamChainSwingA2",
    1650564732: "Parameters/ParamCheek",
    3262488404: "Parameters/ParamCheek02",
    3044192194: "Parameters/ParamCheek03",
    722877025: "Parameters/ParamCheek04",
    1124393604: "Parameters/ParamCheek2",
    872665618: "Parameters/ParamCheek3",
    2120432374: "Parameters/ParamCheekAngry",
    4190688805: "Parameters/ParamCheekAngryL",
    63433542: "Parameters/ParamCheekAngryR",
    570935191: "Parameters/ParamCherryPosX",
    1426126593: "Parameters/ParamCherryPosY",
    3423184571: "Parameters/ParamCherryPosZ",
    2425361543: "Parameters/ParamCherryX",
    796091174: "Parameters/ParamChuckFuwa",
    1088401188: "Parameters/ParamChuckSwing",
    2521177425: "Parameters/ParamChuckSwingH",
    3779145159: "Parameters/ParamChuckSwingI",
    13298184: "Parameters/ParamCloakHuwaB",
    2646264349: "Parameters/ParamCloakSwingA",
    3313808680: "Parameters/ParamClothFuwa1",
    1552802962: "Parameters/ParamClothFuwa2",
    2508309524: "Parameters/ParamClothFuwaA",
    210277806: "Parameters/ParamClothFuwaB",
    2073011512: "Parameters/ParamClothFuwaC",
    3857395867: "Parameters/ParamClothFuwaD",
    39001500: "Parameters/ParamClothFuwaJ",
    716274293: "Parameters/ParamClothHuwaA",
    3015191503: "Parameters/ParamClothHuwaB",
    3300858713: "Parameters/ParamClothHuwaC",
    1524338426: "Parameters/ParamClothHuwaD",
    639921375: "Parameters/ParamClothesFuwa",
    2226631315: "Parameters/ParamClothesSwing",
    1173588212: "Parameters/ParamCloudUpDown",
    2565317455: "Parameters/ParamCloudY",
    1921022208: "Parameters/ParamClowdFuwa",
    3287315499: "Parameters/ParamCollarFuwa",
    1048727694: "Parameters/ParamCollarSwing",
    4282613642: "Parameters/ParamCollarswing",
    1857891283: "Parameters/ParamCollorSwing",
    610714875: "Parameters/ParamCorsetSwing1",
    3178206529: "Parameters/ParamCorsetSwing2",
    2182970360: "Parameters/ParamCostumeFuwa",
    555302639: "Parameters/ParamCostumeSwing",
    955573577: "Parameters/ParamCreamPosX",
    1341380063: "Parameters/ParamCreamPosY",
    3606742117: "Parameters/ParamCreamPosZ",
    2506156731: "Parameters/ParamCreamX",
    619702575: "Parameters/ParamCry",
    1882652453: "Parameters/ParamDarkcycles",
    432666763: "Parameters/ParamDoubleForm",
    1863143841: "Parameters/ParamDressSwing",
    3171936018: "Parameters/ParamEarFuwa",
    811321787: "Parameters/ParamEarHuwaL1",
    2840762369: "Parameters/ParamEarHuwaL2",
    3826972260: "Parameters/ParamEarHuwaR1",
    2098440158: "Parameters/ParamEarHuwaR2",
    3731008440: "Parameters/ParamEarL",
    2894196411: "Parameters/ParamEarLFuwa",
    1854677020: "Parameters/ParamEarLUpDown",
    4023736775: "Parameters/ParamEarLswing1",
    1994131581: "Parameters/ParamEarLswing2",
    611161819: "Parameters/ParamEarR",
    1500366735: "Parameters/ParamEarRUpDown",
    2697710: "Parameters/ParamEarRingFuwa",
    3774242525: "Parameters/ParamEarRingHuwa",
    3578773484: "Parameters/ParamEarRingSwing",
    3627335252: "Parameters/ParamEarRswing1",
    1094578158: "Parameters/ParamEarRswing2",
    1643718421: "Parameters/ParamEarSwing",
    3015659347: "Parameters/ParamEarY",
    1104549151: "Parameters/ParamEarsizeL",
    3151571068: "Parameters/ParamEarsizeR",
    298392022: "Parameters/ParamEarswingL1",
    2294302828: "Parameters/ParamEarswingL2",
    1551988659: "Parameters/ParamEarswingR2",
    1346256061: "Parameters/ParamEnbiFuwa",
    3617001497: "Parameters/ParamEye",
    2422726897: "Parameters/ParamEye0",
    3881881703: "Parameters/ParamEye1",
    2120843741: "Parameters/ParamEye2",
    158249291: "Parameters/ParamEye3",
    2534024424: "Parameters/ParamEye4",
    3758953598: "Parameters/ParamEye5",
    439726873: "Parameters/ParamEye5L",
    3761933946: "Parameters/ParamEye5R",
    2030323140: "Parameters/ParamEye6",
    235107666: "Parameters/ParamEye7",
    2171796666: "Parameters/ParamEyeBallX",
    4134915116: "Parameters/ParamEyeBallY",
    2878588271: "Parameters/ParamEyeChange",
    421411201: "Parameters/ParamEyeEyeLid1",
    3248512125: "Parameters/ParamEyeEyelid01",
    3416285413: "Parameters/ParamEyeHi",
    317062067: "Parameters/ParamEyeHighlight",
    3620524049: "Parameters/ParamEyeHilightonoff",
    832372986: "Parameters/ParamEyeHlForm",
    4097991376: "Parameters/ParamEyeHlOn",
    3989116191: "Parameters/ParamEyeHlOn01",
    1959543973: "Parameters/ParamEyeHlOn02",
    1581384632: "Parameters/ParamEyeKirakira",
    690431522: "Parameters/ParamEyeKirakiraOn",
    795883630: "Parameters/ParamEyeLEyelid",
    2199097593: "Parameters/ParamEyeLForm",
    3728205605: "Parameters/ParamEyeLForm2",
    1978396178: "Parameters/ParamEyeLOpen",
    542740187: "Parameters/ParamEyeLSmile",
    1179028871: "Parameters/ParamEyeLSmileR",
    3641703790: "Parameters/ParamEyeMoth01",
    2714432104: "Parameters/ParamEyeMoth1",
    412213245: "Parameters/ParamEyeREyelid",
    1556330778: "Parameters/ParamEyeRForm",
    3890971854: "Parameters/ParamEyeRForm2",
    2856044529: "Parameters/ParamEyeROpen",
    428207408: "Parameters/ParamEyeRSmile",
    642471233: "Parameters/ParamEyeScale",
    1081240475: "Parameters/ParamEyeSize",
    2867046361: "Parameters/ParamEyeSpecial",
    2120816101: "Parameters/ParamEyeStar",
    1900066077: "Parameters/ParamEyeTeary",
    2040697441: "Parameters/ParamEyeTeary2",
    245736183: "Parameters/ParamEyeTeary3",
    3778377521: "Parameters/ParamEyeWrinkle01L",
    456804946: "Parameters/ParamEyeWrinkle01R",
    3390579954: "Parameters/ParamEyeWrinkle02L",
    806820241: "Parameters/ParamEyeWrinkle02R",
    3176763497: "Parameters/ParamEyeWrinkleL",
    2339494597: "Parameters/ParamEyeWrinkleL01",
    1254580187: "Parameters/ParamEyeWrinkleL01_Form",
    3384875282: "Parameters/ParamEyeWrinkleL01_X",
    3200657796: "Parameters/ParamEyeWrinkleL01_Y",
    309889919: "Parameters/ParamEyeWrinkleL02",
    3428003189: "Parameters/ParamEyeWrinkleL02_Form",
    3414665035: "Parameters/ParamEyeWrinkleL02_X",
    3162544093: "Parameters/ParamEyeWrinkleL02_Y",
    4214242001: "Parameters/ParamEye_kira",
    1393507652: "Parameters/ParamEyeballSwitch",
    1085526974: "Parameters/ParamEyeballX",
    934601512: "Parameters/ParamEyeballY",
    44710175: "Parameters/ParamEyebrowChange",
    3366914530: "Parameters/ParamEyelids",
    1768857238: "Parameters/ParamFaceShadoe2",
    2492426703: "Parameters/ParamFaceShadow",
    295366213: "Parameters/ParamFaceShadow2",
    1979090438: "Parameters/ParamFastenerFuwa",
    2502553909: "Parameters/ParamFastenerHuwa",
    2079001097: "Parameters/ParamFastenerSwing",
    4202413081: "Parameters/ParamFastenerSwing1",
    1668476323: "Parameters/ParamFastenerSwing2",
    3877712832: "Parameters/ParamFastnerSwing",
    3608202235: "Parameters/ParamFinger1",
    1310300737: "Parameters/ParamFinger2",
    1301629155: "Parameters/ParamFinger2_2",
    958311127: "Parameters/ParamFinger3",
    1280778964: "Parameters/ParamFinger3_2",
    2809806708: "Parameters/ParamFinger4",
    1226323025: "Parameters/ParamFinger4_2",
    3497856994: "Parameters/ParamFinger5",
    1222270566: "Parameters/ParamFinger5_2",
    4188325498: "Parameters/ParamFingerL",
    61602585: "Parameters/ParamFingerR",
    1805095925: "Parameters/ParamFrillFuwa",
    3613994765: "Parameters/ParamFrillFuwa2",
    1598146746: "Parameters/ParamFrillSwing",
    826189143: "Parameters/ParamFrillSwing2",
    4164060625: "Parameters/ParamFrillSwingA",
    1631310955: "Parameters/ParamFrillSwingB",
    2986457928: "Parameters/ParamFrontSkirtSwing",
    1784565814: "Parameters/ParamFroshikiHangZ",
    2287383532: "Parameters/ParamFroshikiHuwa",
    2130395650: "Parameters/ParamFroshikiSwing01",
    3891401656: "Parameters/ParamFroshikiSwing02",
    4252684422: "Parameters/ParamFrownLine",
    1158157453: "Parameters/ParamFumeChange",
    4111916540: "Parameters/ParamGlassOn",
    3095524686: "Parameters/ParamHL1Size",
    2936508169: "Parameters/ParamHairAccessory",
    1671823654: "Parameters/ParamHairAccessory2",
    1415909555: "Parameters/ParamHairAccessorySwing1",
    3446423817: "Parameters/ParamHairAccessorySwing2",
    3173065058: "Parameters/ParamHairAho",
    3117719384: "Parameters/ParamHairAho2",
    1293934098: "Parameters/ParamHairAhoge",
    3911307419: "Parameters/ParamHairAhoge2",
    447917413: "Parameters/ParamHairBack",
    2142010931: "Parameters/ParamHairBack01",
    3869585289: "Parameters/ParamHairBack02",
    3189688525: "Parameters/ParamHairBack1",
    655858039: "Parameters/ParamHairBack2",
    1343269345: "Parameters/ParamHairBack3",
    3994794481: "Parameters/ParamHairBackA",
    3498141502: "Parameters/ParamHairBackA2",
    1997727819: "Parameters/ParamHairBackB",
    4222366973: "Parameters/ParamHairBackB2",
    2360018027: "Parameters/ParamHairBackB3",
    334715107: "Parameters/ParamHairBackLr",
    600178327: "Parameters/ParamHairBackLr2",
    3770647160: "Parameters/ParamHairChoro",
    1156315265: "Parameters/ParamHairChoro2",
    303615838: "Parameters/ParamHairControl",
    2078879298: "Parameters/ParamHairFront",
    3520626583: "Parameters/ParamHairFront01",
    1221668397: "Parameters/ParamHairFront02",
    2189195167: "Parameters/ParamHairFront2",
    1265653529: "Parameters/ParamHairFrontA",
    3531146915: "Parameters/ParamHairFrontB",
    4203172644: "Parameters/ParamHairFrontFrizz",
    1018208455: "Parameters/ParamHairFrontHuwa",
    1082396629: "Parameters/ParamHairFuwa",
    2690349286: "Parameters/ParamHairHuwa",
    1394385464: "Parameters/ParamHairHuwa2",
    605786798: "Parameters/ParamHairHuwa3",
    3890600484: "Parameters/ParamHairPartsFuwa",
    640794374: "Parameters/ParamHairPartsFuwa2",
    2920896689: "Parameters/ParamHairPartsSwing",
    1032873597: "Parameters/ParamHairRibbon",
    2335421010: "Parameters/ParamHairRibbonHuwa",
    2212593617: "Parameters/ParamHairRibbonL",
    1921772682: "Parameters/ParamHairRibbonL1",
    3951213872: "Parameters/ParamHairRibbonL2",
    2798310229: "Parameters/ParamHairRibbonR1",
    1069777647: "Parameters/ParamHairRibbonR2",
    1778024379: "Parameters/ParamHairRibbonStringHuwa",
    1425897447: "Parameters/ParamHairSide",
    270528895: "Parameters/ParamHairSide01",
    2299961541: "Parameters/ParamHairSide02",
    3185992987: "Parameters/ParamHairSide1",
    619656353: "Parameters/ParamHairSide2",
    3991082023: "Parameters/ParamHairSideA",
    3207784562: "Parameters/ParamHairSideA2",
    1961510301: "Parameters/ParamHairSideB",
    2485107633: "Parameters/ParamHairSideB2",
    3810036519: "Parameters/ParamHairSideB3",
    4252381227: "Parameters/ParamHairSideFrizz",
    3063165309: "Parameters/ParamHairSideFrizz2",
    185013802: "Parameters/ParamHairSideTail",
    3244650701: "Parameters/ParamHairSideTail2",
    1678048238: "Parameters/ParamHairSwing",
    4260259553: "Parameters/ParamHairTwin",
    3441787525: "Parameters/ParamHairTwin2",
    3215027315: "Parameters/ParamHairTwintailHuwa1",
    648560073: "Parameters/ParamHairTwintailHuwa2",
    1826430306: "Parameters/ParamHairTwintalHuwa1",
    4124331224: "Parameters/ParamHairTwintalHuwa2",
    968573178: "Parameters/ParamHamdShadow",
    1199840847: "Parameters/ParamHamdShadow2",
    813903577: "Parameters/ParamHamdShadow3",
    2790061969: "Parameters/ParamHandA",
    1061529131: "Parameters/ParamHandB",
    3789098979: "Parameters/ParamHandLA",
    2026921561: "Parameters/ParamHandLB",
    265776847: "Parameters/ParamHandLC",
    2444491628: "Parameters/ParamHandLD",
    3870608378: "Parameters/ParamHandLE",
    2143132224: "Parameters/ParamHandLF",
    146451158: "Parameters/ParamHandLG",
    2550511431: "Parameters/ParamHandLH",
    4009920465: "Parameters/ParamHandLI",
    1980487275: "Parameters/ParamHandLJ",
    17622781: "Parameters/ParamHandLK",
    2674422622: "Parameters/ParamHandLL",
    3899605960: "Parameters/ParamHandLM",
    1902514802: "Parameters/ParamHandLN",
    107029220: "Parameters/ParamHandLO",
    2338926353: "Parameters/ParamHandLP",
    4235075463: "Parameters/ParamHandLQ",
    1701269053: "Parameters/ParamHandLR",
    308313771: "Parameters/ParamHandLS",
    2349132552: "Parameters/ParamHandLT",
    4211334046: "Parameters/ParamHandLU",
    899166268: "Parameters/ParamHandRA",
    2895208838: "Parameters/ParamHandRB",
    3684061456: "Parameters/ParamHandRC",
    1173537971: "Parameters/ParamHandRD",
    854979621: "Parameters/ParamHandRE",
    2885461407: "Parameters/ParamHandRF",
    3707475209: "Parameters/ParamHandRG",
    1279559832: "Parameters/ParamHandRH",
    994293774: "Parameters/ParamHandRI",
    2722818484: "Parameters/ParamHandRJ",
    3578648866: "Parameters/ParamHandRK",
    1260997761: "Parameters/ParamHandRL",
    1009679383: "Parameters/ParamHandRM",
    2770808237: "Parameters/ParamHandRN",
    3525319995: "Parameters/ParamHandRO",
    1596461262: "Parameters/ParamHandRP",
    674177112: "Parameters/ParamHandRQ",
    2972086754: "Parameters/ParamHandRR",
    3324068212: "Parameters/ParamHandRS",
    1480971479: "Parameters/ParamHandRT",
    792912961: "Parameters/ParamHandRU",
    913681141: "Parameters/ParamHandRibbonSwing",
    3297800182: "Parameters/ParamHatPosition",
    3901298168: "Parameters/ParamHeadHangZ",
    3764026384: "Parameters/ParamHeadShine",
    503887426: "Parameters/ParamHeadSwingA",
    2265049080: "Parameters/ParamHeadSwingB",
    4026980206: "Parameters/ParamHeadSwingC",
    1851933389: "Parameters/ParamHeadSwingD",
    426078811: "Parameters/ParamHeadSwingE",
    2154570721: "Parameters/ParamHeadSwingF",
    4150989687: "Parameters/ParamHeadSwingG",
    1741950694: "Parameters/ParamHeadSwingH",
    282279536: "Parameters/ParamHeadSwingI",
    2490204078: "Parameters/ParamHeadSwingZ",
    2784609431: "Parameters/ParamHemFuwa",
    4231243417: "Parameters/ParamHemSwing",
    241073746: "Parameters/ParamHighlight",
    2680082603: "Parameters/ParamHighlight2",
    3904495677: "Parameters/ParamHighlight3",
    240225759: "Parameters/ParamHighlightForm",
    2720324382: "Parameters/ParamHighlightView",
    2037511695: "Parameters/ParamHighlightView1",
    3766003637: "Parameters/ParamHighlightView2",
    152765056: "Parameters/ParamHighlightView4",
    1888112791: "Parameters/ParamHimoFuwa",
    591020629: "Parameters/ParamHoodX",
    1413313219: "Parameters/ParamHoodY",
    2859155943: "Parameters/ParamHuwaControl",
    1111428467: "Parameters/ParamInnerSwing",
    4285603815: "Parameters/ParamJacketFuwa",
    3397378595: "Parameters/ParamJacketFuwa02",
    531523796: "Parameters/ParamJacketHuwa",
    1383605806: "Parameters/ParamJacketHuwaA",
    3413211028: "Parameters/ParamJacketHuwaB",
    2892783486: "Parameters/ParamJacketSwing",
    2915339881: "Parameters/ParamJacketSwing2",
    1690873583: "Parameters/ParamJacketSwingA",
    2366728306: "Parameters/ParamJacketYure",
    4031079831: "Parameters/ParamJackettFuwa",
    4234549816: "Parameters/ParamJackettSwing",
    2790893277: "Parameters/ParamJewelryFuwa",
    1970393414: "Parameters/ParamKira",
    471235267: "Parameters/ParamKira1",
    1660276386: "Parameters/ParamKira1Size",
    2233412473: "Parameters/ParamKira2",
    626373746: "Parameters/ParamKira2Size",
    4061666287: "Parameters/ParamKira3",
    406166978: "Parameters/ParamKira3Size",
    697577431: "Parameters/ParamKiraKira",
    1191063520: "Parameters/ParamKiraPin",
    2309390569: "Parameters/ParamKirakira",
    2067958139: "Parameters/ParamKneeRibbonL",
    205896173: "Parameters/ParamKneeRibbonM",
    2941204780: "Parameters/ParamKuchi",
    676255654: "Parameters/ParamKumaView",
    1523262320: "Parameters/ParamLTear",
    2202781583: "Parameters/ParamLaceSwing",
    178000637: "Parameters/ParamLegLR",
    3185367402: "Parameters/ParamLegLRise",
    3930717155: "Parameters/ParamLegLX",
    2639186805: "Parameters/ParamLegLY",
    71801551: "Parameters/ParamLegLZ",
    3739036962: "Parameters/ParamLegRR",
    1645015177: "Parameters/ParamLegRRise",
    1040761916: "Parameters/ParamLegRX",
    1225782442: "Parameters/ParamLegRY",
    3490096400: "Parameters/ParamLegRZ",
    2299297415: "Parameters/ParamLegX",
    4262153745: "Parameters/ParamLegY",
    1728192427: "Parameters/ParamLegZ",
    2371033367: "Parameters/ParamLightView",
    2623341031: "Parameters/ParamLioneye01",
    89411677: "Parameters/ParamLioneye02",
    1918066891: "Parameters/ParamLioneye03",
    3963090280: "Parameters/ParamLioneye04",
    2603664894: "Parameters/ParamLioneye05",
    37327940: "Parameters/ParamLioneye06",
    1967039698: "Parameters/ParamLioneye07",
    3181819922: "Parameters/ParamLooseHair01",
    615508392: "Parameters/ParamLooseHair02",
    2563193013: "Parameters/ParamMaskOn",
    1853511834: "Parameters/ParamMelonPosX",
    427649036: "Parameters/ParamMelonPosY",
    2155100598: "Parameters/ParamMelonPosZ",
    554823962: "Parameters/ParamMelonX",
    2523974014: "Parameters/ParamMouth",
    1139864531: "Parameters/ParamMouth0",
    888652613: "Parameters/ParamMouth1",
    2919142143: "Parameters/ParamMouth2",
    3673793129: "Parameters/ParamMouth3",
    1151154122: "Parameters/ParamMouth4",
    865732444: "Parameters/ParamMouth5",
    2861782758: "Parameters/ParamMouth6",
    3717490288: "Parameters/ParamMouth7",
    3514263446: "Parameters/ParamMouthForm",
    57212699: "Parameters/ParamMouthForm2",
    3995501637: "Parameters/ParamMouthOpenY",
    3165428247: "Parameters/ParamMouthOpenY_old",
    2181955505: "Parameters/ParamMouthPositionX",
    4111126311: "Parameters/ParamMouthPositionY",
    497791140: "Parameters/ParamMouthScaleX",
    1789689906: "Parameters/ParamMouthScaleY",
    894830046: "Parameters/ParamMouthShadow",
    3478896626: "Parameters/ParamMoveFume",
    3201299668: "Parameters/ParamNeckFuwa",
    2658313561: "Parameters/ParamNeckSwin",
    335380611: "Parameters/ParamNeckSwing",
    4152279213: "Parameters/ParamNeckSwing1",
    1853329687: "Parameters/ParamNeckSwing2",
    2053786751: "Parameters/ParamNecklaceFuwa",
    657730868: "Parameters/ParamNecklaceStringHuwaD",
    2495748948: "Parameters/ParamNecklaceStringSwingD",
    3821349826: "Parameters/ParamNecklaceStringSwingE",
    1379753819: "Parameters/ParamNecklaceSwing",
    3862825774: "Parameters/ParamNecklaceSwing2",
    2975251979: "Parameters/ParamNecklaceSwingF",
    3327233693: "Parameters/ParamNecklaceSwingG",
    334399850: "Parameters/ParamNormalEye",
    1064906689: "Parameters/ParamNoseRed",
    3555393053: "Parameters/ParamOrangePosX",
    2767048331: "Parameters/ParamOrangePosY",
    1038393137: "Parameters/ParamOrangePosZ",
    3869781295: "Parameters/ParamOrangeX",
    2251208416: "Parameters/ParamOsageL",
    2082573187: "Parameters/ParamOsageR",
    1218822606: "Parameters/ParamPantsRibbonSwing",
    1715598459: "Parameters/ParamPantsRibbonSwing2",
    4037855955: "Parameters/ParamParkerFuwa",
    276020704: "Parameters/ParamParkerHuwa",
    2379707934: "Parameters/ParamParkerSwing",
    3763819544: "Parameters/ParamParkerSwing2",
    1481285104: "Parameters/ParamPiasSwing",
    1036927352: "Parameters/ParamPiasuSwing",
    3751362731: "Parameters/ParamPierce",
    1527668013: "Parameters/ParamPierce2",
    1461859982: "Parameters/ParamPosition",
    3185949274: "Parameters/ParamPositionX",
    3403844300: "Parameters/ParamPositionY",
    2233141907: "Parameters/ParamRTear",
    3083608617: "Parameters/ParamRedEyeR",
    1728348901: "Parameters/ParamReflectedLight",
    2227145507: "Parameters/ParamRibbon2Swing",
    3090575604: "Parameters/ParamRibbon2Swing2",
    2720634347: "Parameters/ParamRibbonFuwa",
    766487656: "Parameters/ParamRibbonFuwa2",
    1123476184: "Parameters/ParamRibbonHuwa",
    2777107423: "Parameters/ParamRibbonSwing",
    686801017: "Parameters/ParamRibbonSwing01",
    2984677827: "Parameters/ParamRibbonSwing02",
    3336675669: "Parameters/ParamRibbonSwing03",
    1485124854: "Parameters/ParamRibbonSwing04",
    2501224377: "Parameters/ParamRibbonSwing1",
    203224579: "Parameters/ParamRibbonSwing2",
    1763798882: "Parameters/ParamRibbonX",
    1879883664: "Parameters/ParamRibbonXA",
    3909479978: "Parameters/ParamRibbonXB",
    505831412: "Parameters/ParamRibbonY",
    1763151569: "Parameters/ParamRibbonYA",
    4028555115: "Parameters/ParamRibbonYB",
    3711684804: "Parameters/ParamRoprHuwaB",
    263566780: "Parameters/ParamRoprSwingA",
    1424576275: "Parameters/ParamRotationZ",
    3841175661: "Parameters/ParamScale",
    2377047131: "Parameters/ParamScarf1",
    346525153: "Parameters/ParamScarf2",
    2022373968: "Parameters/ParamShimenawaFaceSwing",
    1761407924: "Parameters/ParamShimenawaHuwa",
    1985753683: "Parameters/ParamShimenawaRingSwing",
    3660462105: "Parameters/ParamShimenawaSwing01",
    1126656419: "Parameters/ParamShimenawaSwing02",
    167184000: "Parameters/ParamShirtFuwa",
    407962998: "Parameters/ParamShirtHuwaB",
    1867920864: "Parameters/ParamShirtHuwaC",
    3004587123: "Parameters/ParamShirtSwingB",
    3289476325: "Parameters/ParamShirtSwingC",
    2437270373: "Parameters/ParamShirtsFuwa",
    2947844518: "Parameters/ParamShirtsSwing",
    2009336143: "Parameters/ParamShoulder01L",
    2378904620: "Parameters/ParamShoulder01R",
    1558796940: "Parameters/ParamShoulder02L",
    2800117743: "Parameters/ParamShoulder02R",
    151499700: "Parameters/ParamShoulderL",
    4077424343: "Parameters/ParamShoulderR",
    551282230: "Parameters/ParamShyView",
    795402279: "Parameters/ParamSize",
    3220202374: "Parameters/ParamSize2",
    3900401720: "Parameters/ParamSkeleton0",
    2675726510: "Parameters/ParamSkeleton1",
    108333332: "Parameters/ParamSkeleton2",
    1903311234: "Parameters/ParamSkeleton3",
    4011240481: "Parameters/ParamSkeleton4",
    2443564567: "Parameters/ParamSkirt",
    2156391276: "Parameters/ParamSkirt2Swing",
    1219303786: "Parameters/ParamSkirtBertSwing",
    84257039: "Parameters/ParamSkirtFrillHuwa",
    3178758115: "Parameters/ParamSkirtFrillSwing",
    2272788835: "Parameters/ParamSkirtFuwa",
    1474340692: "Parameters/ParamSkirtFuwa1",
    3471439598: "Parameters/ParamSkirtFuwa2",
    3119433336: "Parameters/ParamSkirtFuwa3",
    663376859: "Parameters/ParamSkirtFuwa4",
    1739028048: "Parameters/ParamSkirtHuwa",
    568080819: "Parameters/ParamSkirtHuwaB",
    2766679669: "Parameters/ParamSkirtLaceHuwa",
    218532422: "Parameters/ParamSkirtLaceSwing",
    1075815090: "Parameters/ParamSkirtRibbonFuwa",
    3553727041: "Parameters/ParamSkirtShake",
    1187058009: "Parameters/ParamSkirtSwing",
    3465061400: "Parameters/ParamSkirtSwing01",
    1468134818: "Parameters/ParamSkirtSwing02",
    2435696630: "Parameters/ParamSkirtSwing1",
    136615500: "Parameters/ParamSkirtSwing2",
    2133042906: "Parameters/ParamSkirtSwing3",
    3779539833: "Parameters/ParamSkirtSwing4",
    3240670922: "Parameters/ParamSkirtSwingA",
    1478616944: "Parameters/ParamSkirtSwingB",
    791074790: "Parameters/ParamSkirtSwingC",
    2973909573: "Parameters/ParamSkirtSwingD",
    3326440147: "Parameters/ParamSkirtSwingE",
    1167521556: "Parameters/ParamSkirtSwingR",
    848414594: "Parameters/ParamSkirtSwingS",
    2901758497: "Parameters/ParamSkirtSwingT",
    2602036815: "Parameters/ParamSleeveFuwa",
    2076744060: "Parameters/ParamSleeveHuwa",
    1946489429: "Parameters/ParamSleeveSwing",
    3055370847: "Parameters/ParamSleeveSwingL",
    1276280636: "Parameters/ParamSleeveSwingR",
    171939558: "Parameters/ParamSmile",
    1046190462: "Parameters/ParamSmokeForm01",
    2807220420: "Parameters/ParamSmokeForm02",
    39356985: "Parameters/ParamSmokeSwing02",
    1857169349: "Parameters/ParamSmokeView01",
    4156257919: "Parameters/ParamSmokeView02",
    2472867533: "Parameters/ParamSpeakerResize",
    213325773: "Parameters/ParamSpecialCheek01",
    3342417859: "Parameters/ParamSpecialEye",
    2582661823: "Parameters/ParamSpecialEye01",
    16325381: "Parameters/ParamSpecialEye02",
    2013146003: "Parameters/ParamSpecialEye03",
    410981432: "Parameters/ParamSpecialEye2",
    177352564: "Parameters/ParamSpecialEyeRotate1",
    1715562423: "Parameters/ParamSpecialEyebrow01",
    3728752230: "Parameters/ParamStar1PositionX",
    2840006384: "Parameters/ParamStar1PositionY",
    2610605975: "Parameters/ParamStar1Rotation",
    800879294: "Parameters/ParamStar1Scaling",
    897025381: "Parameters/ParamStar2PositionX",
    1114682867: "Parameters/ParamStar2PositionY",
    2719454034: "Parameters/ParamStar2Rotation",
    2704504157: "Parameters/ParamStar2Scaling",
    3669356123: "Parameters/ParamStar3PositionX",
    2914180813: "Parameters/ParamStar3PositionY",
    3043788561: "Parameters/ParamStar3Rotation",
    1838773699: "Parameters/ParamStar3Scaling",
    2197501062: "Parameters/ParamStarAllonoff",
    3216432596: "Parameters/ParamStaronoff",
    3586705073: "Parameters/ParamStrawberryPosX",
    2731537959: "Parameters/ParamStrawberryPosY",
    1002874781: "Parameters/ParamStrawberryPosZ",
    231510313: "Parameters/ParamStrawberryX",
    2828308470: "Parameters/ParamStringFuwa",
    3331037724: "Parameters/ParamStringSwing",
    971818426: "Parameters/ParamStringSwing01",
    2699400192: "Parameters/ParamStringSwing02",
    2535143522: "Parameters/ParamStringSwing1",
    236087768: "Parameters/ParamStringSwing2",
    141487747: "Parameters/ParamSugaoSwing",
    1852572281: "Parameters/ParamSugaoSwing2",
    2639993062: "Parameters/ParamSuitHuwa",
    1681521550: "Parameters/ParamSuitSwing",
    391498689: "Parameters/ParamSweaMove2",
    4046985140: "Parameters/ParamSweat1",
    1748059662: "Parameters/ParamSweat2",
    910715833: "Parameters/ParamSweatMove",
    3872502433: "Parameters/ParamSweatMoveon",
    88717352: "Parameters/ParamSweatOn",
    795172687: "Parameters/ParamSweatOn2",
    663484239: "Parameters/ParamSweatView",
    4237388283: "Parameters/ParamSweatView2",
    4282366110: "Parameters/ParamSweatonoff",
    2170085279: "Parameters/ParamSwingA",
    407997989: "Parameters/ParamSwingB",
    1867955891: "Parameters/ParamSwingC",
    2732879099: "Parameters/ParamSwingCardigan",
    2819788061: "Parameters/ParamSwingCoat",
    4046596880: "Parameters/ParamSwingD",
    2251627398: "Parameters/ParamSwingE",
    524044860: "Parameters/ParamSwingF",
    1748728490: "Parameters/ParamSwingG",
    4169426747: "Parameters/ParamSwingH",
    3805815803: "Parameters/ParamSwingHeadA",
    3708339165: "Parameters/ParamSwingHeadHachimaki",
    541857961: "Parameters/ParamSwingHeadribbon",
    2407749549: "Parameters/ParamSwingI",
    378144279: "Parameters/ParamSwingJ",
    1192411332: "Parameters/ParamSwingJacket",
    1636644481: "Parameters/ParamSwingK",
    2142601779: "Parameters/ParamSwingKatyusha",
    2776263321: "Parameters/ParamSwingKatyusha2",
    388770144: "Parameters/ParamTailSwing",
    1467020156: "Parameters/ParamTailSwing2",
    494375721: "Parameters/ParamTasukiHuwaB",
    1484411378: "Parameters/ParamTasukiHuwaB2",
    3519964986: "Parameters/ParamTasukiSwingA",
    3691555288: "Parameters/ParamTasukiSwingA2",
    2436919207: "Parameters/ParamTear",
    1391904809: "Parameters/ParamTear2",
    3629100631: "Parameters/ParamTear2on",
    636614847: "Parameters/ParamTear3",
    3147132188: "Parameters/ParamTear4",
    1798597939: "Parameters/ParamTearHighlightL",
    2436583504: "Parameters/ParamTearHighlightR",
    3846935570: "Parameters/ParamTearL",
    3918103761: "Parameters/ParamTearL2",
    2506119870: "Parameters/ParamTearLine",
    524593521: "Parameters/ParamTearR",
    1036565262: "Parameters/ParamTearR2",
    905834095: "Parameters/ParamTearon",
    2283503513: "Parameters/ParamTeeth",
    2943939027: "Parameters/ParamTeeth01",
    913416297: "Parameters/ParamTeeth02",
    2475593362: "Parameters/ParamTeeth2",
    3834232324: "Parameters/ParamTeeth3",
    3126364231: "Parameters/ParamTemp01",
    1110817175: "Parameters/ParamTieFuwa",
    4241779428: "Parameters/ParamTieSwing",
    561217570: "Parameters/ParamTwintail",
    3461348719: "Parameters/ParamTwintails",
    3340143929: "Parameters/ParamTwintails2",
    2953952687: "Parameters/ParamTwintails3",
    1118730002: "Parameters/ParamTwintailsFuwa",
    3912130866: "Parameters/ParamTwintailsFuwa2",
    2409744776: "Parameters/ParamTwintailsHangZ",
    4146016103: "Parameters/ParamTwintailsLTwist",
    3468905612: "Parameters/ParamTwintailsRTwist",
    1490014742: "Parameters/ParamTwintailsTwistL",
    2730553205: "Parameters/ParamTwintailsTwistR",
    2127276225: "Parameters/ParamUnderBodyHangZ",
    1513497407: "Parameters/ParamUpperBody",
    4018162809: "Parameters/ParamUpperBodyZ",
    2444971114: "Parameters/ParamVestSwing",
    3269570882: "Parameters/ParamWaistcord1Swing01",
    1541988600: "Parameters/ParamWaistcord1Swing02",
    201222549: "Parameters/ParamWaistcord2ySwing",
    650216962: "Parameters/ParamWhiskerSwingL",
    3704535905: "Parameters/ParamWhiskerSwingR",
    921748034: "Parameters/ParamWorkRotate",
    39434875: "Parameters/ParamWrinkle",
    2074752843: "Parameters/ParamZipperFuwa",
    1938320277: "Parameters/ParamZipperSwing",
    854485227: "Parameters/Param_Angle_Rotation10",
    1172920445: "Parameters/Param_Angle_Rotation11",
    3705670087: "Parameters/Param_Angle_Rotation12",
    2884057425: "Parameters/Param_Angle_Rotation13",
    897821938: "Parameters/Param_Angle_Rotation14",
    1115987044: "Parameters/Param_Angle_Rotation15",
    3683503582: "Parameters/Param_Angle_Rotation16",
    2894789960: "Parameters/Param_Angle_Rotation17",
    1010163929: "Parameters/Param_Angle_Rotation18",
    1261621327: "Parameters/Param_Angle_Rotation19",
    432225064: "Parameters/Param_Angle_Rotation20",
    1858341822: "Parameters/Param_Angle_Rotation21",
    4157430276: "Parameters/Param_Angle_Rotation22",
    514785073: "Parameters/Param_Angle_Rotation24",
    1772736423: "Parameters/Param_Angle_Rotation25",
    4037058077: "Parameters/Param_Angle_Rotation26",
    2275913355: "Parameters/Param_Angle_Rotation27",
    387494682: "Parameters/Param_Angle_Rotation28",
    1612678028: "Parameters/Param_Angle_Rotation29",
    14158441: "Parameters/Param_Angle_Rotation30",
    2011118335: "Parameters/Param_Angle_Rotation31",
    4007029573: "Parameters/Param_Angle_Rotation32",
    2580634579: "Parameters/Param_Angle_Rotation33",
    129355376: "Parameters/Param_Angle_Rotation34",
    1890778854: "Parameters/Param_Angle_Rotation35",
    3921391452: "Parameters/Param_Angle_Rotation36",
    2663161802: "Parameters/Param_Angle_Rotation37",
    235111003: "Parameters/Param_Angle_Rotation38",
    2030351053: "Parameters/Param_Angle_Rotation39",
    1335467182: "Parameters/Param_Angle_Rotation40",
    1223973047: "Parameters/Param_Angle_Rotation44",
    1072916513: "Parameters/Param_Angle_Rotation45",
    2801416603: "Parameters/Param_Angle_Rotation46",
    3523021069: "Parameters/Param_Angle_Rotation47",
    1094849692: "Parameters/Param_Angle_Rotation48",
    910500874: "Parameters/Param_Angle_Rotation49",
    1451404783: "Parameters/Param_Angle_Rotation50",
    562404729: "Parameters/Param_Angle_Rotation51",
    3096235203: "Parameters/Param_Angle_Rotation52",
    2802511257: "Parameters/Param_Angle_Rotation6",
    3490569487: "Parameters/Param_Angle_Rotation7",
    1085456542: "Parameters/Param_Angle_Rotation8",
    934670344: "Parameters/Param_Angle_Rotation9",
    2910755435: "Parameters/Param_Angle_RotationL_10_ArtMesh181",
    920376836: "Parameters/Param_Angle_RotationL_11_ArtMesh181",
    1211087100: "Parameters/Param_Angle_RotationL_1_ArtMesh181",
    1068591628: "Parameters/Param_Angle_RotationL_2_ArtMesh181",
    2752783971: "Parameters/Param_Angle_RotationL_3_ArtMesh181",
    3498889196: "Parameters/Param_Angle_RotationL_4_ArtMesh181",
    1261011843: "Parameters/Param_Angle_RotationL_5_ArtMesh181",
    1018646899: "Parameters/Param_Angle_RotationL_6_ArtMesh181",
    2802982172: "Parameters/Param_Angle_RotationL_7_ArtMesh181",
    3582362221: "Parameters/Param_Angle_RotationL_8_ArtMesh181",
    1310969346: "Parameters/Param_Angle_RotationL_9_ArtMesh181",
    1028406218: "Parameters/Param_Angle_Rotation_10_ArtMesh181",
    3276456759: "Parameters/Param_Angle_Rotation_10_ArtMesh325",
    759472667: "Parameters/Param_Angle_Rotation_10_ArtMesh327",
    3711710831: "Parameters/Param_Angle_Rotation_10_ArtMesh331",
    2800318373: "Parameters/Param_Angle_Rotation_11_ArtMesh181",
    1492094808: "Parameters/Param_Angle_Rotation_11_ArtMesh325",
    3068259956: "Parameters/Param_Angle_Rotation_11_ArtMesh327",
    1184463360: "Parameters/Param_Angle_Rotation_11_ArtMesh331",
    3246341252: "Parameters/Param_Angle_Rotation_12_ArtMesh327",
    822585584: "Parameters/Param_Angle_Rotation_12_ArtMesh331",
    1524265195: "Parameters/Param_Angle_Rotation_13_ArtMesh327",
    2862802079: "Parameters/Param_Angle_Rotation_13_ArtMesh331",
    776116580: "Parameters/Param_Angle_Rotation_14_ArtMesh327",
    3728346384: "Parameters/Param_Angle_Rotation_14_ArtMesh331",
    3051869451: "Parameters/Param_Angle_Rotation_15_ArtMesh327",
    1168064895: "Parameters/Param_Angle_Rotation_15_ArtMesh331",
    3262711803: "Parameters/Param_Angle_Rotation_16_ArtMesh327",
    838964111: "Parameters/Param_Angle_Rotation_16_ArtMesh331",
    1507609492: "Parameters/Param_Angle_Rotation_17_ArtMesh327",
    2846154720: "Parameters/Param_Angle_Rotation_17_ArtMesh331",
    726193381: "Parameters/Param_Angle_Rotation_18_ArtMesh327",
    3677372561: "Parameters/Param_Angle_Rotation_18_ArtMesh331",
    2968361098: "Parameters/Param_Angle_Rotation_19_ArtMesh327",
    1083510014: "Parameters/Param_Angle_Rotation_19_ArtMesh331",
    3314780403: "Parameters/Param_Angle_Rotation_1_ArtMesh181",
    999649294: "Parameters/Param_Angle_Rotation_1_ArtMesh325",
    3583709474: "Parameters/Param_Angle_Rotation_1_ArtMesh327",
    841169957: "Parameters/Param_Angle_Rotation_1_ArtMesh329",
    635667798: "Parameters/Param_Angle_Rotation_1_ArtMesh331",
    2987219459: "Parameters/Param_Angle_Rotation_2_ArtMesh181",
    1275831038: "Parameters/Param_Angle_Rotation_2_ArtMesh325",
    2718286802: "Parameters/Param_Angle_Rotation_2_ArtMesh327",
    1170075349: "Parameters/Param_Angle_Rotation_2_ArtMesh329",
    1383946150: "Parameters/Param_Angle_Rotation_2_ArtMesh331",
    698881644: "Parameters/Param_Angle_Rotation_3_ArtMesh181",
    3618565777: "Parameters/Param_Angle_Rotation_3_ArtMesh325",
    966823869: "Parameters/Param_Angle_Rotation_3_ArtMesh327",
    3726156474: "Parameters/Param_Angle_Rotation_3_ArtMesh329",
    3386383305: "Parameters/Param_Angle_Rotation_3_ArtMesh331",
    1563492323: "Parameters/Param_Angle_Rotation_4_ArtMesh181",
    2738226974: "Parameters/Param_Angle_Rotation_4_ArtMesh325",
    1295541810: "Parameters/Param_Angle_Rotation_4_ArtMesh327",
    2860534581: "Parameters/Param_Angle_Rotation_4_ArtMesh329",
    3175150150: "Parameters/Param_Angle_Rotation_4_ArtMesh331",
    3331699596: "Parameters/Param_Angle_Rotation_5_ArtMesh181",
    949177201: "Parameters/Param_Angle_Rotation_5_ArtMesh325",
    3600624221: "Parameters/Param_Angle_Rotation_5_ArtMesh327",
    824510298: "Parameters/Param_Angle_Rotation_5_ArtMesh329",
    652590633: "Parameters/Param_Angle_Rotation_5_ArtMesh331",
    2970313084: "Parameters/Param_Angle_Rotation_6_ArtMesh181",
    1326291329: "Parameters/Param_Angle_Rotation_6_ArtMesh325",
    2701392045: "Parameters/Param_Angle_Rotation_6_ArtMesh327",
    1186715050: "Parameters/Param_Angle_Rotation_6_ArtMesh329",
    1367043289: "Parameters/Param_Angle_Rotation_6_ArtMesh331",
    716057875: "Parameters/Param_Angle_Rotation_7_ArtMesh181",
    3567834606: "Parameters/Param_Angle_Rotation_7_ArtMesh325",
    984003778: "Parameters/Param_Angle_Rotation_7_ArtMesh327",
    3709770181: "Parameters/Param_Angle_Rotation_7_ArtMesh329",
    3403554998: "Parameters/Param_Angle_Rotation_7_ArtMesh331",
    1480241762: "Parameters/Param_Angle_Rotation_8_ArtMesh181",
    2788969119: "Parameters/Param_Angle_Rotation_8_ArtMesh325",
    1211247539: "Parameters/Param_Angle_Rotation_8_ArtMesh327",
    2945058484: "Parameters/Param_Angle_Rotation_8_ArtMesh329",
    3091906503: "Parameters/Param_Angle_Rotation_8_ArtMesh331",
    3282042381: "Parameters/Param_Angle_Rotation_9_ArtMesh181",
    1033438960: "Parameters/Param_Angle_Rotation_9_ArtMesh325",
    3549915100: "Parameters/Param_Angle_Rotation_9_ArtMesh327",
    875514587: "Parameters/Param_Angle_Rotation_9_ArtMesh329",
    602928040: "Parameters/Param_Angle_Rotation_9_ArtMesh331",
    2745217400: "Parameters/Paramcheek",
    905596179: "Parameters/Paramhand01",
    2901613737: "Parameters/Paramhand02",
    2360343793: "Parameters/Paramhighlight",
    513109378: "Parameters/Paramkiraonoff",
    1809445648: "Parameters/ParamshirtHuwa",
    806432332: "Parameters/ParamshirtSwing",
    2464978475: "Parameters/Paramtears",
    496017766: "Parameters/Paramtongue",
    100476007: "Parameters/uruuru",
}



import zipfile
import UnityPy
import logging
import json
import sys

from io import BytesIO

from UnityPy.enums import ClassIDType
from UnityPy.helpers.Tpk import get_typetree_node
from sssekai.unity.AssetBundle import load_assetbundle
from tqdm import tqdm

REGION_MAP = {
    # NOTE: Order is used to determine the region
    "com.hermes.mk": "cn",
    "com.hermes.mk.asia": "tw",
    "com.sega.ColorfulStage.en": "en",
    "com.sega.pjsekai": "jp",
    "com.pjsekai.kr": "kr",
}
logger = logging.getLogger("apphash")


def enum_candidates(zip_file, filter):
    return (
        (f, zip_file.open(f), zip_file) for f in zip_file.filelist if filter(f.filename)
    )


def enum_package(zip_file):
    yield zip_file
    for f in zip_file.filelist:
        if f.filename.lower().endswith(".apk"):
            yield zipfile.ZipFile(zip_file.open(f))


def dump_axml_stringpool(f: BytesIO):
    read_int = lambda nbytes: int.from_bytes(f.read(nbytes), "little")
    f.seek(8)
    hdr_type, hdr_size, size = read_int(2), read_int(2), read_int(4)
    n_strings = read_int(4)
    unk = read_int(4)
    flags = read_int(4)
    is_utf8 = flags & (1 << 8)
    string_offset = read_int(4) + 8
    unk = read_int(4)
    offsets = [read_int(4) for _ in range(n_strings)]
    for offset in offsets:
        f.seek(string_offset + offset)
        string_len = read_int(2)
        if is_utf8:
            string = f.read(string_len).decode("utf-8")
        else:
            string = f.read(string_len * 2).decode("utf-16-le")
        yield string


def main_apphash(args):
    env = UnityPy.Environment()
    app_package = None
    app_version = "unknown"
    app_hash = "unknown"

    if not args.ab_src:
        if not args.apk_src or args.fetch:
            from requests import get

            src = BytesIO()
            logger.debug("Fetching latest game package (JP) from APKPure")
            resp = get(
                "https://d.apkpure.net/b/XAPK/com.sega.pjsekai?version=latest",
                stream=True,
            )
            size = resp.headers.get("Content-Length", -1)
            with tqdm(total=int(size), unit="B", unit_scale=True) as progress:
                for chunck in resp.iter_content(chunk_size=2**20):
                    src.write(chunck)
                    progress.update(len(chunck))
            src.seek(0)
        else:
            src = open(args.apk_src, "rb")
        with zipfile.ZipFile(src, "r") as zip_ref:
            manifests = [
                manifest
                for package in enum_package(zip_ref)
                for manifest in enum_candidates(
                    package, lambda fn: fn == "AndroidManifest.xml"
                )
            ]
            manifest = manifests[0][1]
            manifest_strings = set(dump_axml_stringpool(manifest))
            # Heur: Reverse lookup the package name from the manifest strings
            for ky in REGION_MAP:
                if ky in manifest_strings:
                    app_package = ky
                    break
            candidates = [
                candidate
                for package in enum_package(zip_ref)
                for candidate in enum_candidates(
                    package,
                    lambda fn: fn.split("/")[-1]
                    in {
                        "6350e2ec327334c8a9b7f494f344a761",  # PJSK Android
                        "c726e51b6fe37463685916a1687158dd",  # PJSK iOS
                        "data.unity3d",  # TW,KR,CN (ByteDance)
                    },
                )
            ]
            for candidate, stream, _ in candidates:
                env.load_file(stream)
    else:
        logger.info("Loading from AssetBundle %s" % args.ab_src)
        with open(args.ab_src, "rb") as f:
            env = load_assetbundle(BytesIO(f.read()))

    from sssekai.generated import UTTCGen_AsInstance
    from sssekai.generated.Sekai import (
        AndroidPlayerSettingConfig,
        IOSPlayerSettingConfig,
    )

    for reader in env.objects:
        if reader.type == UnityPy.enums.ClassIDType.MonoBehaviour:
            pname = reader.peek_name()
            clazz = {
                "production_android": AndroidPlayerSettingConfig,
                "production_ios": IOSPlayerSettingConfig,
            }
            clazz = clazz.get(pname, None)
            if clazz:
                # Works with post 3.4 (JP 3rd Anniversary) builds and downstream regional builds
                config = UTTCGen_AsInstance(clazz, reader)
                config: AndroidPlayerSettingConfig | IOSPlayerSettingConfig
                app_version = "%s.%s.%s" % (
                    config.clientMajorVersion,
                    config.clientMinorVersion,
                    config.clientBuildVersion,
                )
                data_version = "%s.%s.%s" % (
                    config.clientDataMajorVersion,
                    config.clientDataMinorVersion,
                    config.clientDataBuildVersion,
                )
                ab_version = "%s.%s.%s" % (
                    config.clientMajorVersion,
                    config.clientMinorVersion,
                    config.clientDataRevision,
                )
                app_hash = config.clientAppHash
                # XXX: CN build has the same package name as the JP build??
                # We need another heuristic to determine the package name
                # This is previously done by setting app_package
                package_heur = app_package or config.bundleIdentifier
                region = region = REGION_MAP.get(package_heur, "unknown")
                print(
                    f"Found {config.productName} at {config.m_Name}",
                    f"  Memo: {config.memo}",
                    f"  Package: {config.bundleIdentifier} (actually assumed as {package_heur})",
                    f"  AppHash (app_hash):     {config.clientAppHash}",
                    f"  Region  (app_region):   {region} (determined by {package_heur})",
                    f"  Version (app_version):  {app_version}",
                    f"  Version (ab_version):   {ab_version}",
                    f"  Bundle Version: {config.bundleVersion}",
                    f"  Data Version:   {data_version}",
                    f"  Version Suffix: {config.clientVersionSuffix}",
                    "",
                    "",
                    sep="\n",
                    file=sys.stderr,
                )
                # Only keep the Android one
                if pname == "production_android":
                    app_package = (
                        app_package or "Unknown Package (Failed APK Heuristic)"
                    )
                    # fmt: off
                    match args.format:
                        case 'json':
                            print(json.dumps({
                                "package": app_package,
                                "reported_package": config.bundleIdentifier,
                                "app_hash": app_hash,
                                "app_region": region,
                                "app_version": app_version,
                                "ab_version": ab_version,
                            }, indent=4))
                        case "markdown":
                            print(f"""{app_package} ({app_version}, {region})
---
Reported Package: {config.bundleIdentifier}

|{'app_hash'.rjust(48)}|   app_region|  app_version|   ab_version|
|{   '-'.rjust(48,'-')}|-------------|-------------|-------------|
|{  app_hash.rjust(48)}|{region.rjust(13)}|{app_version.rjust(13)}|{ab_version.rjust(13)}|

- CLI Usage:

        sssekai abcache --app-platform android --app-region {region} --app-version {app_version} --app-appHash {app_hash} --app-abVersion {ab_version}

- Python Usage:

        from sssekai.abcache import AbCacheConfig

        AbCacheConfig(
            app_region="{region}",
            app_version="{app_version}",
            ab_version="{ab_version}",
            app_hash="{app_hash}",
            app_platform="android"
        )
""")
